#include "z80.h"
#include "Snapshots.h"
#include "Tape.h"
#include <string.h>

#include <ctype.h>
#include "ULA.h"
#include "Screen.h"
#include "AY.h"
#include "Rom.h"
#include "HostMachine.h"

CZ80 *CPU;
CULA *ULA;
CAY *SND;
CScreen *SCR;
CTape *Tape;

#define BOARDLOC 0xc8c8
#define BOARDLEN 0x78
void LoadScoreBoard()
{
	FILE *sb = NULL;//Resourcefopen("scores.dat", "rb");
	if(sb)
	{
		Uint8 ScoreBoard[BOARDLEN];
		fread(ScoreBoard, 1, BOARDLEN, sb);
		fclose(sb);

		int c = BOARDLEN;
		while(c--)
			CPU->WriteMemory(BOARDLOC+c, ScoreBoard[c]);
	}
}

void SaveScoreBoard()
{
	Uint8 ScoreBoard[BOARDLEN];
	int c = BOARDLEN;
	while(c--)
		CPU->ReadMemory(BOARDLOC+c, ScoreBoard[c]);

	FILE *sb = Resourcefopen("scores.dat", "wb");
	if(sb)
	{
		fwrite(ScoreBoard, 1, BOARDLEN, sb);
		fclose(sb);
	}
}

CULA::~CULA()
{
	while(NumSticks--)
	{
		SDL_JoystickClose(Sticks[NumSticks]);
	}
	delete[] Sticks;
}

CULA::CULA()
{
	Bit6 = 0;
	PageEnable = true;
	EnterSet = false;
	Sticks = NULL;

	if(NumSticks = SDL_NumJoysticks())
	{
		Sticks = new SDL_Joystick *[NumSticks];
		int c = NumSticks;
		while(c--)
		{
			Sticks[c] = SDL_JoystickOpen(c);
		}
	}
}

void CULA::SetEnter(bool r)
{
	EnterSet = r;
}

void CULA::Write(CZ80 &CPU, Uint16 Addr, Uint8 Value, Uint64 Value64, unsigned int Timestamp)
{
	/* write takes only FE and paging writes, so no need for much logic */
	if(!(Addr&1))
	{
		Bit6 = 0x40 & (((Value&0x08) << 3) | ((Value&0x10) << 2));
		SCR->SetBorderCol(Value&7, Timestamp);

		/* pass on speaker level */
		SND->Write(CPU, CAY_TOGGLEADDR, Value&0x10, 0, Timestamp);
		return;
	}

	if(PageEnable)
	{
		CPU.Page(3, 2 + (Value&7), 2 + (Value&7));
		CPU.Page(0, (Value >> 4)&1, 10);
		CPU.SetCopyMask((2+((Value&8) ? 7 : 5)) << 14);

		if(Value&0x20) PageEnable = false;
	}
}

#define JOY_THRESHOLD 16384

void CULA::Read(CZ80 &CPU, Uint16 Addr, Uint8 &Value, Uint64 &Value64, unsigned int Timestamp)
{
	if(!(Addr&1))
	{
		Value = 0xbf | Bit6 | (Tape->ReadBit(Timestamp) ? 0x40 : 0x00);

		Uint8 *Keyboard = SDL_GetKeyState(NULL);

		int JoyButtonState = 0;
		Sint16 JoyYAxis = 0, JoyXAxis = 0;
		SDL_JoystickUpdate();

		int c = NumSticks;
		while(c--)
		{
			JoyXAxis += SDL_JoystickGetAxis(Sticks[c], 0);
			JoyYAxis += SDL_JoystickGetAxis(Sticks[c], 1);

			int TotalButtons = SDL_JoystickNumButtons(Sticks[c]);
			int buttons = (TotalButtons > 4) ? 4 : TotalButtons;
			while(buttons--)
			{
				if(SDL_JoystickGetButton(Sticks[c], buttons))
					JoyButtonState |= 1 << (buttons&1);
			}
		}

		if(!(Addr&0x100))
		{
			if(Keyboard[SDLK_RSHIFT] | Keyboard[SDLK_LSHIFT] || Keyboard[SDLK_LEFT] || Keyboard[SDLK_RIGHT] || Keyboard[SDLK_DOWN] || Keyboard[SDLK_UP] || Keyboard[SDLK_BACKSPACE])
				Value &= ~0x1;
			if(Keyboard[SDLK_z])	Value &= ~0x2;
			if(Keyboard[SDLK_x])	Value &= ~0x4;
			if(Keyboard[SDLK_c])	Value &= ~0x8;
			if(Keyboard[SDLK_v])	Value &= ~0x10;
		}
		if(!(Addr&0x200))
		{
			if(Keyboard[SDLK_a] || (JoyYAxis > JOY_THRESHOLD))	Value &= ~0x1;
			if(Keyboard[SDLK_s])	Value &= ~0x2;
			if(Keyboard[SDLK_d])	Value &= ~0x4;
			if(Keyboard[SDLK_f])	Value &= ~0x8;
			if(Keyboard[SDLK_g])	Value &= ~0x10;
		}
		if(!(Addr&0x400))
		{	
			if(Keyboard[SDLK_q] || (JoyYAxis < -JOY_THRESHOLD) || (JoyButtonState&2))	Value &= ~0x1;
			if(Keyboard[SDLK_w])	Value &= ~0x2;
			if(Keyboard[SDLK_e])	Value &= ~0x4;
			if(Keyboard[SDLK_r])	Value &= ~0x8;
			if(Keyboard[SDLK_t])	Value &= ~0x10;
		}
		if(!(Addr&0x800))
		{
			if(Keyboard[SDLK_1])	Value &= ~0x1;
			if(Keyboard[SDLK_2])	Value &= ~0x2;
			if(Keyboard[SDLK_3])	Value &= ~0x4;
			if(Keyboard[SDLK_4])	Value &= ~0x8;
			if(Keyboard[SDLK_5] || Keyboard[SDLK_LEFT])	Value &= ~0x10;
		}
		if(!(Addr&0x1000))
		{
			if(Keyboard[SDLK_0] || Keyboard[SDLK_BACKSPACE])	Value &= ~0x1;
			if(Keyboard[SDLK_9])	Value &= ~0x2;
			if(Keyboard[SDLK_8] || Keyboard[SDLK_RIGHT])	Value &= ~0x4;
			if(Keyboard[SDLK_7] || Keyboard[SDLK_UP])	Value &= ~0x8;
			if(Keyboard[SDLK_6] || Keyboard[SDLK_DOWN])	Value &= ~0x10;
		}
		if(!(Addr&0x2000))
		{
			if(Keyboard[SDLK_p] || (JoyXAxis > JOY_THRESHOLD))	Value &= ~0x1;
			if(Keyboard[SDLK_o] || (JoyXAxis < -JOY_THRESHOLD))	Value &= ~0x2;
			if(Keyboard[SDLK_i])	Value &= ~0x4;
			if(Keyboard[SDLK_u])	Value &= ~0x8;
			if(Keyboard[SDLK_y])	Value &= ~0x10;
		}
		if(!(Addr&0x4000))
		{
			if(Keyboard[SDLK_RETURN] || EnterSet)
				Value &= ~0x1;
			if(Keyboard[SDLK_l])	Value &= ~0x2;
			if(Keyboard[SDLK_k])	Value &= ~0x4;
			if(Keyboard[SDLK_j])	Value &= ~0x8;
			if(Keyboard[SDLK_h])	Value &= ~0x10;
		}
		if(!(Addr&0x8000))
		{
			if(Keyboard[SDLK_SPACE] || (JoyButtonState&1))	Value &= ~0x1;
			if(Keyboard[SDLK_RCTRL] || Keyboard[SDLK_LCTRL] || Keyboard[SDLK_COMMA] || Keyboard[SDLK_PERIOD])	Value &= ~0x2;
			if(Keyboard[SDLK_m] || Keyboard[SDLK_PERIOD])	Value &= ~0x4;
			if(Keyboard[SDLK_n] || Keyboard[SDLK_COMMA])	Value &= ~0x8;
			if(Keyboard[SDLK_b])	Value &= ~0x10;
		}
		return;
	}

	/* otherwise return last ULA byte - screen, attr or whatever */
	Value = 0xff;
}

class CKempston : public CZ80IO
{
	public:
		void Write(CZ80 &CPU, Uint16 Addr, Uint8 Value8, Uint64 Value64, Uint32 Time);
		void Read(CZ80 &CPU, Uint16 Addr, Uint8 &Value8, Uint64 &Value64, Uint32 Time);
};
void CKempston::Write(CZ80 &CPU, Uint16 Addr, Uint8 Value8, Uint64 Value64, Uint32 Time)
{
}

void CKempston::Read(CZ80 &CPU, Uint16 Addr, Uint8 &Value8, Uint64 &Value64, Uint32 Time)
{
	Value8 = 0xff;
}

#define DanStartAddr	0xd100
#define DanLength		488
Uint64 DanShapes[DanLength*3];

void TrapMultiplexer(void *p, CZ80 &CPU, Uint16 Addr)
{
	int R, W;
	switch(Addr)
	{
		case 0x056c: Tape->TapeTrap(CPU, Addr); break;
		case 0xb11e:
			/* backup Dan sprites from 0xd136 */
			CPU.GetPaged(DanStartAddr >> 14, R, W);
			memcpy(DanShapes, &CPU.GetPtr64(R)[DanStartAddr&16383], DanLength*3*sizeof(Uint64));
		break;

		case 0xb15e:
			/* flip sets of three bytes from 37147 to +488*3 */
			Uint64 V1, V2, V3;
			int c;
			for(c = 0; c < 160; c++)
			{
				V1 = DanShapes[(c*3) + 0];
				V2 = DanShapes[(c*3) + 1];
				V3 = DanShapes[(c*3) + 2];

#define Reverse(v)\
				v =\
					((v&0x00000000000000ffull) << 56) |\
					((v&0x000000000000ff00ull) << 40) |\
					((v&0x0000000000ff0000ull) << 24) |\
					((v&0x00000000ff000000ull) << 8) |\
					((v&0x000000ff00000000ull) >> 8) |\
					((v&0x0000ff0000000000ull) >> 24) |\
					((v&0x00ff000000000000ull) >> 40) |\
					((v&0xff00000000000000ull) >> 56)

				Reverse(V1);
				Reverse(V2);
				Reverse(V3);

				DanShapes[(c*3) + 0] = V3;
				DanShapes[(c*3) + 1] = V2;
				DanShapes[(c*3) + 2] = V1;
			}
			for(c = 192; c < DanLength; c++)
			{
				V1 = DanShapes[(c*3) + 0];
				V2 = DanShapes[(c*3) + 1];
				V3 = DanShapes[(c*3) + 2];

				Reverse(V1);
				Reverse(V2);
				Reverse(V3);

				DanShapes[(c*3) + 0] = V3;
				DanShapes[(c*3) + 1] = V2;
				DanShapes[(c*3) + 2] = V1;
			}

			CPU.GetPaged(DanStartAddr >> 14, R, W);
			memcpy(&CPU.GetPtr64(R)[DanStartAddr&16383], DanShapes, DanLength*3*sizeof(Uint64));
		break;
	}
}

Uint8 Palette[256][3];
int PalTarget;

void ClearPalette()
{
	PalTarget = 1;
}

void InstallPalette()
{
	const Uint8 *SrcPalette = SCR->GetMultiPalette();
	Uint8 DstPalette[256*129*3];
	int c = 129;
	while(c--)
	{
		int forecolour = (c&7) | (c >> 7);
		int ic = 256;
		while(ic--)
		{
			DstPalette[(c*256 + ic)*3 + 0] = forecolour ? Palette[ic][0] : 0;
			DstPalette[(c*256 + ic)*3 + 1] = forecolour ? Palette[ic][1] : 0;
			DstPalette[(c*256 + ic)*3 + 2] = forecolour ? Palette[ic][2] : 0;
		}

		if(c < 128)
		{
			DstPalette[c*256*3 + 0] = SrcPalette[c*256*3 + 0];
			DstPalette[c*256*3 + 1] = SrcPalette[c*256*3 + 1];
			DstPalette[c*256*3 + 2] = SrcPalette[c*256*3 + 2];

			int ic = 8;
			while(ic--)
			{
				DstPalette[(c*256 + 248 + ic)*3 + 0] = ((ic+1)*SrcPalette[(c*256 + 255)*3 + 0] + (7-ic)*SrcPalette[c*256*3 + 0]) / 8;
				DstPalette[(c*256 + 248 + ic)*3 + 1] = ((ic+1)*SrcPalette[(c*256 + 255)*3 + 1] + (7-ic)*SrcPalette[c*256*3 + 1]) / 8;
				DstPalette[(c*256 + 248 + ic)*3 + 2] = ((ic+1)*SrcPalette[(c*256 + 255)*3 + 2] + (7-ic)*SrcPalette[c*256*3 + 2]) / 8;
			}
		}
	}

	SCR->SetMultiPalette(DstPalette);
}

int GetIndex(Uint8 r, Uint8 g, Uint8 b)
{
	for(int c = 1; c < PalTarget; c++)
	{
		if(	(Palette[c][0] == r) &&
			(Palette[c][1] == g) &&
			(Palette[c][2] == b) )
			return c;
	}

	if(PalTarget < 248)
	{
		Palette[PalTarget][0] = r;
		Palette[PalTarget][1] = g;
		Palette[PalTarget][2] = b;
		PalTarget++;
	}

	return PalTarget-1;
}

int InstallGraphicalResource(char *name, int loc, int strips = 1)
{
	SDL_Surface *Sprites;
	
	char *resname = GetResourceName(name);
	Sprites = SDL_LoadBMP(resname);
	delete [] resname;

	if(!Sprites) return loc-1;
	int BWidth = (Sprites->w / strips) >> 3;

	SDL_LockSurface(Sprites);
	/* remap colours here */
	int x, y;
	for(y = 0; y < Sprites->h; y++)
		for(x = 0; x < Sprites->w; x++)
		{
			Uint8 index = ((Uint8 *)Sprites->pixels)[ (y*Sprites->pitch) + x];
			if((index) && (index < 248))
			{
				SDL_Color *col = &Sprites->format->palette->colors[ index ];
				((Uint8 *)Sprites->pixels)[ (y*Sprites->pitch) + x] = GetIndex(col->r, col->g, col->b);
			}
		}

	int xp = 0;
	for(int xl = 0; xl < Sprites->w; xl += Sprites->w/strips)
	{
		y = Sprites->h;
		while(y--)
		{
			for(x = 0; x < BWidth; x++)
			{
				Uint64 Colour;

				Colour =
					((Uint64)(((Uint8 *)Sprites->pixels)[ (y*Sprites->pitch) + ((x) << 3) + 7 +xl]) << 0) |
					((Uint64)(((Uint8 *)Sprites->pixels)[ (y*Sprites->pitch) + ((x) << 3) + 6 +xl]) << 8) |
					((Uint64)(((Uint8 *)Sprites->pixels)[ (y*Sprites->pitch) + ((x) << 3) + 5 +xl]) << 16) |
					((Uint64)(((Uint8 *)Sprites->pixels)[ (y*Sprites->pitch) + ((x) << 3) + 4 +xl]) << 24) |
					((Uint64)(((Uint8 *)Sprites->pixels)[ (y*Sprites->pitch) + ((x) << 3) + 3 +xl]) << 32) |
					((Uint64)(((Uint8 *)Sprites->pixels)[ (y*Sprites->pitch) + ((x) << 3) + 2 +xl]) << 40) |
					((Uint64)(((Uint8 *)Sprites->pixels)[ (y*Sprites->pitch) + ((x) << 3) + 1 +xl]) << 48) |
					((Uint64)(((Uint8 *)Sprites->pixels)[ (y*Sprites->pitch) + ((x) << 3) + 0 +xl]) << 56);
				Uint8 ival;
				int Addr = loc + xp + (y*BWidth) + x;
				CPU->ReadMemory(Addr , ival);
				CPU->WriteMemory(Addr, ival, Colour);
			}
		}
		xp += Sprites->h;
	}
	SDL_UnlockSurface(Sprites);
	int res = loc + Sprites->h*BWidth;

	/* and integrate palette */
	/*const Uint8 *SrcPalette = SCR->GetMultiPalette();
	Uint8 DstPalette[256*129*3];
	int c = 129;
	while(c--)
	{
		int forecolour = (c&7) | (c >> 7);
		int ic = 256;
		while(ic--)
		{
			DstPalette[(c*256 + ic)*3 + 0] = forecolour ? Sprites->format->palette->colors[ic].r : 0;
			DstPalette[(c*256 + ic)*3 + 1] = forecolour ? Sprites->format->palette->colors[ic].g : 0;
			DstPalette[(c*256 + ic)*3 + 2] = forecolour ? Sprites->format->palette->colors[ic].b : 0;
		}

		if(c < 128)
		{
			DstPalette[c*256*3 + 0] = SrcPalette[c*256*3 + 0];
			DstPalette[c*256*3 + 1] = SrcPalette[c*256*3 + 1];
			DstPalette[c*256*3 + 2] = SrcPalette[c*256*3 + 2];

			int ic = 8;
			while(ic--)
			{
				DstPalette[(c*256 + 248 + ic)*3 + 0] = ((ic+1)*SrcPalette[(c*256 + 255)*3 + 0] + (7-ic)*SrcPalette[c*256*3 + 0]) / 8;
				DstPalette[(c*256 + 248 + ic)*3 + 1] = ((ic+1)*SrcPalette[(c*256 + 255)*3 + 1] + (7-ic)*SrcPalette[c*256*3 + 1]) / 8;
				DstPalette[(c*256 + 248 + ic)*3 + 2] = ((ic+1)*SrcPalette[(c*256 + 255)*3 + 2] + (7-ic)*SrcPalette[c*256*3 + 2]) / 8;
			}
		}
	}

	SCR->SetMultiPalette(DstPalette);*/

	SDL_FreeSurface(Sprites);

	return res;
}

void DanTrap(void *a, CZ80 &b, Uint16 Addr, Uint16 Length)
{
	if(Addr == 0x6400)
	{
		/* set gfx mode to multiplexed, remove trap function */
		SCR->SetFlags(SCR->GetFlags() | CSF_MULTIPLEXED);
		Tape->SetTrapFunction(NULL);

		/* load score data */
		LoadScoreBoard();

		/* load improved graphics */
		ClearPalette();

		/* tiles first */
		InstallGraphicalResource("tiles.bmp", 42307 + 16384 - 27, 14);

		/* now sprites */
		int eaddr;
		eaddr = InstallGraphicalResource("dan.bmp", DanStartAddr);
		eaddr = InstallGraphicalResource("treens1.bmp", eaddr);
		eaddr = InstallGraphicalResource("treens2.bmp", eaddr);
		eaddr = InstallGraphicalResource("treens3.bmp", eaddr);
		eaddr = InstallGraphicalResource("mekon.bmp", eaddr);
		InstallGraphicalResource("bombs.bmp", eaddr);

		/* palette, finally */
		InstallPalette();

		/* install dan flipping trap */
		CPU->SetTrapAddr(0xb11e, true);
		CPU->SetTrapAddr(0xb15e, true);
		CPU->SetTrapFunction(TrapMultiplexer, NULL);
	}
}

#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif

int main(int argc, char *argv[])
{
	CKempston joy;
	bool quit = false;
	bool DanDare = false;

	int c = 2, *b = &c, **a = &b;
	int *&av = *(int **)a;
	av=b;
	b = av;

	SDL_Init(SDL_INIT_VIDEO | SDL_INIT_AUDIO | SDL_INIT_JOYSTICK);
	SetupHostMachine();
	
	CPU = new CZ80;
	ULA = new CULA;
	SCR = new CScreen;
	SND = new CAY;
	Tape = new CTape(ULA);

	if(!SCR->HasScreen())
	{
#ifdef WIN32
		MessageBox(NULL, "Could not initiate display", "Dan Dare: Fatal Error", MB_OK);
#else
		fprintf(stderr, "Could not initiate display\n");
#endif
	}
	else
	{
		CPU->AllocateMemory(11);

		CPU->SetIOReadDevice(0x0001, 0x0000, ULA);
		CPU->SetIOContention(0x0001, 0x0000, true);	/* ULA is contended */

		CPU->SetIOWriteDevice(0x0001, 0x0000, ULA);
		CPU->SetIOWriteDevice(0x8002, 0x0000, ULA);
		CPU->SetIOReadDevice(0x00ff, 0x001f, &joy);

		/* install ROM */
		WriteROM(CPU->GetPtr8(0));

		/* multi up from ROM */
		CPU->DumpMultiplexedCopy();
		CPU->SetTrapFunction(Tape->TapeTrap, Tape);
		CPU->SetTrapAddr(0x056c, true);

		SCR->Attach(CPU);
		SCR->SetFlags(0);

		CPU->Page(0, 0, 10);
		CPU->Page(1, 5+2, 5+2);
		CPU->Page(2, 2+2, 2+2);
		CPU->Page(3, 4+2, 4+2);

		CPU->SetIOReadDevice(0xffff, 0xfffd, SND);
		CPU->SetIOWriteDevice(0xffff, 0xbffd, SND);
		CPU->SetIOWriteDevice(0xffff, 0xfffd, SND);

		if(argc > 1)
		{
			char name[4];
			strcpy(name, argv[1] + strlen(argv[1]) - 3);
			
			/* Mac OS X seems to have no stricmp (?) */
			int c = 3;
			while(c--) name[c] = tolower(name[c]);

			if(!strcmp(name, "sna")) LoadSNA(*CPU, argv[1], 0);
			if(!strcmp(name, "z80")) LoadZ80(*CPU, argv[1], 0);
			if(!strcmp(name, "tap") || !strcmp(name, "tzx")) if(Tape->Open(argv[1], 0)) ULA->SetEnter(true);
		}
		else
		{
			char *TapeName = GetResourceName("DanDare1.tap");
			if(Tape->Open(TapeName, 0))
			{
				ULA->SetEnter(true);
				Tape->SetTrapFunction(DanTrap);
				DanDare = true;
			}
			delete[] TapeName;
		}

		Uint32 FrameEnd = SDL_GetTicks() + 20;

		bool behind = false;
		while(!quit)
		{
			CPU->Update((228*311) - 32, behind);
			CPU->SetIRQ(true);
			SCR->Update(228*311, behind);
			SND->Update(228*311, behind);
			CPU->Update(32, behind);
			CPU->SetIRQ(false);

			Uint32 SleepPeriod = SDL_GetTicks();
			SleepPeriod = FrameEnd - SleepPeriod;

			if(SleepPeriod < 20)
			{
				SDL_Delay(SleepPeriod);
				FrameEnd += 20;
				behind = false;
			}
			else
			{
				FrameEnd = SDL_GetTicks()+20;
				behind = true;
			}

			SDL_Event ev;
			while(SDL_PollEvent(&ev))
			{
				switch(ev.type)
				{
					case SDL_QUIT: quit = true; break;

					case SDL_KEYDOWN:
						/* MacOS style key shortcuts */
						if(ev.key.keysym.mod&KMOD_META)
							switch(ev.key.keysym.sym)
							{
								default: break;
								case SDLK_q: quit = true; break;
								case SDLK_f: SCR->ToggleFullscreen(); break;
							}

						/* Windows style key shortcuts */
						if(ev.key.keysym.mod&(KMOD_LALT|KMOD_RALT))
						{
							switch(ev.key.keysym.sym)
							{
								case SDLK_F4: quit = true; break;
								case SDLK_RETURN: SCR->ToggleFullscreen(); break;
								default: break;
							}
						}

						/* overlay toggler */
						if(ev.key.keysym.sym == SDLK_F1)
						{
							SCR->SetFlags(SCR->GetFlags()^CSF_ALLOWOVERLAY);
						}
					break;

/*					case SDL_USEREVENT:
						switch(ev.user.code)
						{
							// Load a File
							case GUIEVT_LOADFILE:
							{
//								PPool->Open( (char *)ev.user.data1 );
								free( ev.user.data1 );
							}
							break;
						}
					break;*/
				}
			}
		}

		Tape->Close();

		if(DanDare)
			SaveScoreBoard();
	}

	delete CPU;
	delete ULA;
	delete SND;
	delete SCR;
	delete Tape;

	return 0;
}
