#include "Screen.h"
#include <memory.h>
#include "z80.h"
#include "HostMachine.h"

#define CONTENDED_FROM_IRQ	14361
#define IRQ_LENGTH			32

#define	CONTENDED_START		(CONTENDED_FROM_IRQ - IRQ_LENGTH)

const Uint8 RTable[16] = {0,   0, 203, 203,   0,   0, 203, 203,   0,   0, 255, 255,   0,   0, 255, 255};
const Uint8 GTable[16] = {0,   0,   0,   0, 203, 203, 203, 203,   0,   0,   0,   0, 255, 255, 255, 255};
const Uint8 BTable[16] = {0, 203,   0, 203,   0, 203,   0, 203,   0, 255,   0, 255,   0, 255,   0, 255};

void CScreen::SetBorderCol(Uint8 Colour, Uint32 Time)
{
	BorderQueue[BWritePtr].Colour = Colour;
	if(Colour == 1)
		Colour++;

	Uint32 Diff = Time - CONTENDED_START;
	Diff = (Diff+3)&~3;
	BorderQueue[BWritePtr].Time = Diff + CONTENDED_START;
	BWritePtr = (BWritePtr+1)&1023;
}

Uint32 CScreen::Update(Uint32 cycles, bool behind)
{
	FrameNo++;
	if(FlagsDirty)
	{
		SetupGraphics();
		FlagsDirty = false;
	}

	if(behind)
	{
		BehindNo++;
		if(BehindNo == 8)
		{
			BehindNo = 0;
			behind = false;
		}
	}

	if(!behind)
	{
		int CopyLen, OffMul;
		Uint8 *Pixels;
		Uint16 Pitch;

		if(over)
		{
			if(SDL_LockYUVOverlay(over)) return 0;
			CopyLen = 16;
			OffMul = 2;
			Pixels = (Uint8 *)over->pixels[0];
			Pitch = over->pitches[0];
		}
		else
		{
			if(SDL_LockSurface(front)) return 0;
			CopyLen = front->format->BytesPerPixel << 3;
			OffMul = front->format->BytesPerPixel;
			Pixels = (Uint8 *)front->pixels;
			Pitch = front->pitch;
		}

		int PelAddr = CONTENDED_START >> 1;
		int y;

		/* draw border */
		Uint32 LTime = Time;

		/* border changes that occured above the top of the displayed area */
		while((BReadPtr != BWritePtr) && (BorderQueue[BReadPtr].Time - Time) < (CONTENDED_START - 4*4 - 24*228))
		{
			BColour = BorderQueue[BReadPtr].Colour;
			BReadPtr = (BReadPtr+1)&1023;
		}
		LTime += (CONTENDED_START - 24*228 - 4*4);

		y = 0;
		int x = 0;
		while(y < 240)
		{
			int TargetBytes;

			if((BReadPtr == BWritePtr) || (BorderQueue[BReadPtr].Time - LTime) > 228)
				TargetBytes = 57 - x;
			else
			{
				TargetBytes = (BorderQueue[BReadPtr].Time - LTime) >> 2;

				if(TargetBytes > (57-x))
					TargetBytes = 57-x;
			}

			int x1, x2;
			x1 = x;
			x2 = x+TargetBytes;
			LTime += TargetBytes << 2;

			if((y < 24) || (y >= 216))
			{
				if(x1 < 40)
				{
					int w = x2 - x1;
					if(x2 > 40)
						w = 40 - x1;

					while(w--)
					{
						memcpy(&Pixels[ ((y << 1)*Pitch) + (x1 << 4)*OffMul], &Lookup[(BColour << 10)*OffMul], CopyLen);
						memcpy(&Pixels[ ((y << 1)*Pitch) + ((x1 << 4) + 8)*OffMul], &Lookup[(BColour << 10)*OffMul], CopyLen);

						memcpy(&Pixels[ (((y << 1)+1)*Pitch) + (x1 << 4)*OffMul], &Lookup[(BColour << 10)*OffMul], CopyLen);
						memcpy(&Pixels[ (((y << 1)+1)*Pitch) + ((x1 << 4) + 8)*OffMul], &Lookup[(BColour << 10)*OffMul], CopyLen);
						x1++;
					}
				}
			}
			else
			{
				if(x1 < 4)
				{
					int w = x2 - x1;
					if(x2 > 4)
						w = 4 - x1;

					while(w--)
					{
						memcpy(&Pixels[ ((y << 1)*Pitch) + (x1 << 4)*OffMul], &Lookup[(BColour << 10)*OffMul], CopyLen);
						memcpy(&Pixels[ ((y << 1)*Pitch) + ((x1 << 4) + 8)*OffMul], &Lookup[(BColour << 10)*OffMul], CopyLen);

						memcpy(&Pixels[ (((y << 1)+1)*Pitch) + (x1 << 4)*OffMul], &Lookup[(BColour << 10)*OffMul], CopyLen);
						memcpy(&Pixels[ (((y << 1)+1)*Pitch) + ((x1 << 4) + 8)*OffMul], &Lookup[(BColour << 10)*OffMul], CopyLen);
						x1++;
					}
				}

				if((x2 > 36) && (x1 < 40))
				{
					int w = x2 - x1;
					if(x2 > 40)
						w = 40 - x1;
					if(x1 < 36)
					{
						w -= 36-x1;
						x1 = 36;
					}

					while(w--)
					{
						memcpy(&Pixels[ ((y << 1)*Pitch) + (x1 << 4)*OffMul], &Lookup[(BColour << 10)*OffMul], CopyLen);
						memcpy(&Pixels[ ((y << 1)*Pitch) + ((x1 << 4) + 8)*OffMul], &Lookup[(BColour << 10)*OffMul], CopyLen);

						memcpy(&Pixels[ (((y << 1)+1)*Pitch) + (x1 << 4)*OffMul], &Lookup[(BColour << 10)*OffMul], CopyLen);
						memcpy(&Pixels[ (((y << 1)+1)*Pitch) + ((x1 << 4) + 8)*OffMul], &Lookup[(BColour << 10)*OffMul], CopyLen);
						x1++;
					}
				}
			}

			if(BorderQueue[BReadPtr].Time == LTime)
			{
				BColour = BorderQueue[BReadPtr].Colour;
				BReadPtr = (BReadPtr+1)&1023;
			}


			x = x2;
			if(x == 57)
			{
				x = 0;
				y++;
			}
		}

		/* pixel area */
		if(Multi)
		{
			switch(front->format->BytesPerPixel)
			{
				case 1:{
					Uint8 *PelPtr = &Pixels[ (48*Pitch) + 64];
					for(y = 24; y < 216; y++)
					{
						for(int x = 0; x < 32; x++)
						{
							Uint64 c = PelData64[PelAddr];
							int attr = PelData8[PelAddr+1];
							PelAddr+=2;

							if((attr&0x80) && (FrameNo&0x10))
								attr = (attr&0x40) | ((attr >> 3)&7) | ((attr&7) << 3);

							/* work pixelwise */
							attr <<= 8;
							int ic = 8;
							while(ic--)
							{
								Uint32 TabPos = (Uint32)(c >> 56) | attr;
								PelPtr[0] = PelPtr[1] = PelPtr[Pitch] = PelPtr[Pitch+1] = MultiLookup[TabPos];
								PelPtr += 4;
								c <<= 8;
							}
						}

						PelAddr += 50;
						PelPtr += (Pitch << 1) - 512;
 					}
				}break;

				case 2:{
					if(over)
					{
						int YOffs1, YOffs2, UOffs, VOffs;
						switch(over->format)
						{
							default:
							case SDL_YVYU_OVERLAY: YOffs1 = 0; YOffs2 = 2; VOffs = 1; UOffs = 3; break;
							case SDL_YUY2_OVERLAY: YOffs1 = 0; YOffs2 = 2; VOffs = 3; UOffs = 1; break;
							case SDL_UYVY_OVERLAY: YOffs1 = 1; YOffs2 = 3; UOffs = 0; VOffs = 2; break;
						}

						
						Uint8 *PelPtr = (Uint8 *)&Pixels[ (48*Pitch) + 128];
						for(y = 24; y < 216; y++)
						{
							for(int x = 0; x < 32; x++)
							{
								Uint64 c = PelData64[PelAddr];
								int attr = PelData8[PelAddr+1];
								PelAddr+=2;

								if((attr&0x80) && (FrameNo&0x10))
									attr = (attr&0x40) | ((attr >> 3)&7) | ((attr&7) << 3);

								/* work bipixelwise */
								int ic = 4;
								while(ic--)
								{
									Uint32 TabPos1 = ((Uint32)(c >> 56) | (attr << 8)) << 2;
									Uint32 TabPos2 = ((Uint32)((c >> 48)&0xff) | (attr << 8)) << 2;

									PelPtr[YOffs1] = MultiLookup[TabPos1];
									PelPtr[YOffs2] = MultiLookup[TabPos2];
									PelPtr[UOffs] = (MultiLookup[TabPos1+1] + MultiLookup[TabPos2+1]) >> 1;
									PelPtr[VOffs] = (MultiLookup[TabPos1+2] + MultiLookup[TabPos2+2]) >> 1;

									PelPtr += 4;
									c <<= 16;
								}
							}

							PelAddr += 50;
							PelPtr += (Pitch << 1) - 512;
						}
					}
					else
					{
						Uint16 *PelPtr = (Uint16 *)&Pixels[ (48*Pitch) + 128];
						for(y = 24; y < 216; y++)
						{
							for(int x = 0; x < 32; x++)
							{
								Uint64 c = PelData64[PelAddr];
								int attr = PelData8[PelAddr+1];
								PelAddr+=2;

								if((attr&0x80) && (FrameNo&0x10))
									attr = (attr&0x40) | ((attr >> 3)&7) | ((attr&7) << 3);

								/* work pixelwise */
								attr <<= 8;
								int ic = 8;
								while(ic--)
								{
									Uint32 TabPos = (Uint32)(c >> 56) | attr;
									PelPtr[0] = PelPtr[1] = PelPtr[Pitch >> 1] = PelPtr[(Pitch >> 1)+1] = ((Uint16 *)MultiLookup)[TabPos];
									PelPtr += 2;
									c <<= 8;
								}
							}

							PelAddr += 50;
							PelPtr += Pitch - 512;
						}
 					}
				}break;

				case 3:{
					Uint8 *PelPtr = &Pixels[ (48*Pitch) + 192];
					for(y = 24; y < 216; y++)
					{
						for(int x = 0; x < 32; x++)
						{
							Uint64 c = PelData64[PelAddr];
							int attr = PelData8[PelAddr+1];
							PelAddr+=2;

							if((attr&0x80) && (FrameNo&0x10))
								attr = (attr&0x40) | ((attr >> 3)&7) | ((attr&7) << 3);

							/* work pixelwise */
							attr <<= 8;
							int ic = 8;
							while(ic--)
							{
								Uint8 *TabPtr = &MultiLookup[((Uint32)(c >> 56) | attr)*3];
								
								PelPtr[0] = PelPtr[3] = PelPtr[Pitch] = PelPtr[Pitch+3] = TabPtr[0];
								PelPtr[1] = PelPtr[4] = PelPtr[Pitch+1] = PelPtr[Pitch+4] = TabPtr[1];
								*PelPtr++ = TabPtr[2];
								PelPtr += 6;
								c <<= 8;
							}
						}

						PelAddr += 50;
						PelPtr += (Pitch << 1) - 1536;
 					}
				}break;

				case 4:{
					Uint32 *PelPtr = (Uint32 *)&Pixels[ (48*Pitch) + 256];
					for(y = 24; y < 216; y++)
					{
						for(int x = 0; x < 32; x++)
						{
							Uint64 c = PelData64[PelAddr];
							int attr = PelData8[PelAddr+1];
							PelAddr+=2;

							if((attr&0x80) && (FrameNo&0x10))
								attr = (attr&0x40) | ((attr >> 3)&7) | ((attr&7) << 3);

							/* work pixelwise */
							attr <<= 8;
							int ic = 8;
							while(ic--)
							{
								Uint32 TabPos = (Uint32)(c >> 56) | attr;
								PelPtr[1] = PelPtr[0] = PelPtr[(Pitch >> 2)] = PelPtr[(Pitch >> 2)+1] = ((Uint32 *)MultiLookup)[TabPos];
								PelPtr+=2;
								c <<= 8;
							}
						}

						PelAddr += 50;
						PelPtr += (Pitch>>1) - 512;
 					}
				}break;
			}
		}
		else
		{
			switch(front->format->BytesPerPixel)
			{
				default: {
					Uint8 *PelPtr1 = &Pixels[ (48*Pitch) + 64*OffMul];
					Uint8 *PelPtr2 = &Pixels[ (49*Pitch) + 64*OffMul];
					for(y = 24; y < 216; y++)
					{
						for(int x = 0; x < 32; x++)
						{
							int c = PelData8[PelAddr], attr = PelData8[PelAddr+1], composite; PelAddr+=2;

							if((attr&0x80) && (FrameNo&0x10))
								attr = (attr&0x40) | ((attr >> 3)&7) | ((attr&7) << 3);
							attr = (attr&0x7f) << 7;

							/* left four pixels */
							composite = ((c&0xf0) >> 1) | attr;
							memcpy(PelPtr1, &Lookup[composite*OffMul], CopyLen);
							PelPtr1 += CopyLen;
							memcpy(PelPtr2, &Lookup[composite*OffMul], CopyLen);
							PelPtr2 += CopyLen;
					
							/* right four pixels */
							composite = ((c&0x0f) << 3) | attr;
							memcpy(PelPtr1, &Lookup[composite*OffMul], CopyLen);
							PelPtr1 += CopyLen;
							memcpy(PelPtr2, &Lookup[composite*OffMul], CopyLen);
							PelPtr2 += CopyLen;
						}

						PelAddr += 50;
						PelPtr1 += (Pitch<<1) - (CopyLen << 6);
						PelPtr2 += (Pitch<<1) - (CopyLen << 6);
					}
				} break;
				
				case 1:
				{
					Uint32 *PelPtr1 = (Uint32 *)&Pixels[ (48*Pitch) + 64*OffMul];
					Uint32 *PelPtr2 = (Uint32 *)&Pixels[ (49*Pitch) + 64*OffMul];
					for(y = 24; y < 216; y++)
					{
						for(int x = 0; x < 32; x++)
						{
							int c = PelData8[PelAddr], attr = PelData8[PelAddr+1]; PelAddr+=2;

							if((attr&0x80) && (FrameNo&0x10))
								attr = (attr&0x40) | ((attr >> 3)&7) | ((attr&7) << 3);
							attr = (attr&0x7f) << 7;

							/* left four pixels */
							Uint32 *TablePtr = (Uint32 *)&Lookup[((c&0xf0) >> 1) | attr];
							PelPtr1[0] = PelPtr2[0] = TablePtr[0];
							PelPtr1[1] = PelPtr2[1] = TablePtr[1];
							PelPtr1 += 2; PelPtr2 += 2;
					
							/* right four pixels */
							TablePtr = (Uint32 *)&Lookup[((c&0x0f) << 3) | attr];
							PelPtr1[0] = PelPtr2[0] = TablePtr[0];
							PelPtr1[1] = PelPtr2[1] = TablePtr[1];
							PelPtr1 += 2; PelPtr2 += 2;
						}

						PelAddr += 50;
						PelPtr1 += (Pitch >> 1) - 128;
						PelPtr2 += (Pitch >> 1) - 128;
					}
				} break;

				case 2:
				{
					Uint32 *PelPtr1 = (Uint32 *)&Pixels[ (48*Pitch) + 64*OffMul];
					Uint32 *PelPtr2 = (Uint32 *)&Pixels[ (49*Pitch) + 64*OffMul];
					for(y = 24; y < 216; y++)
					{
						for(int x = 0; x < 32; x++)
						{
							int c = PelData8[PelAddr], attr = PelData8[PelAddr+1]; PelAddr+=2;

							if((attr&0x80) && (FrameNo&0x10))
								attr = (attr&0x40) | ((attr >> 3)&7) | ((attr&7) << 3);
							attr = (attr&0x7f) << 7;

							/* left four pixels */
							Uint32 *TablePtr = (Uint32 *)&Lookup[(((c&0xf0) >> 1) | attr) << 1];
							PelPtr1[0] = PelPtr2[0] = TablePtr[0];
							PelPtr1[1] = PelPtr2[1] = TablePtr[1];
							PelPtr1[2] = PelPtr2[2] = TablePtr[2];
							PelPtr1[3] = PelPtr2[3] = TablePtr[3];
							PelPtr1 += 4; PelPtr2 += 4;
					
							/* right four pixels */
							TablePtr = (Uint32 *)&Lookup[(((c&0x0f) << 3) | attr) << 1];
							PelPtr1[0] = PelPtr2[0] = TablePtr[0];
							PelPtr1[1] = PelPtr2[1] = TablePtr[1];
							PelPtr1[2] = PelPtr2[2] = TablePtr[2];
							PelPtr1[3] = PelPtr2[3] = TablePtr[3];
							PelPtr1 += 4; PelPtr2 += 4;
						}

						PelAddr += 50;
						PelPtr1 += (Pitch >> 1) - 256;
						PelPtr2 += (Pitch >> 1) - 256;
					}
				} break;

				case 4:
				{
					Uint32 *PelPtr1 = (Uint32 *)&Pixels[ (48*Pitch) + 64*OffMul];
					Uint32 *PelPtr2 = (Uint32 *)&Pixels[ (49*Pitch) + 64*OffMul];
					for(y = 24; y < 216; y++)
					{
						for(int x = 0; x < 32; x++)
						{
							int c = PelData8[PelAddr], attr = PelData8[PelAddr+1]; PelAddr+=2;

							if((attr&0x80) && (FrameNo&0x10))
								attr = (attr&0x40) | ((attr >> 3)&7) | ((attr&7) << 3);
							attr = (attr&0x7f) << 7;

							/* left four pixels */
							Uint32 *TablePtr = (Uint32 *)&Lookup[(((c&0xf0) >> 1) | attr) << 2];
							PelPtr1[0] = PelPtr2[0] = TablePtr[0];
							PelPtr1[1] = PelPtr2[1] = TablePtr[1];
							PelPtr1[2] = PelPtr2[2] = TablePtr[2];
							PelPtr1[3] = PelPtr2[3] = TablePtr[3];
							PelPtr1[4] = PelPtr2[4] = TablePtr[4];
							PelPtr1[5] = PelPtr2[5] = TablePtr[5];
							PelPtr1[6] = PelPtr2[6] = TablePtr[6];
							PelPtr1[7] = PelPtr2[7] = TablePtr[7];
							PelPtr1 += 8; PelPtr2 += 8;
					
							/* right four pixels */
							TablePtr = (Uint32 *)&Lookup[(((c&0x0f) << 3) | attr) << 2];
							PelPtr1[0] = PelPtr2[0] = TablePtr[0];
							PelPtr1[1] = PelPtr2[1] = TablePtr[1];
							PelPtr1[2] = PelPtr2[2] = TablePtr[2];
							PelPtr1[3] = PelPtr2[3] = TablePtr[3];
							PelPtr1[4] = PelPtr2[4] = TablePtr[4];
							PelPtr1[5] = PelPtr2[5] = TablePtr[5];
							PelPtr1[6] = PelPtr2[6] = TablePtr[6];
							PelPtr1[7] = PelPtr2[7] = TablePtr[7];
							PelPtr1 += 8; PelPtr2 += 8;
						}

						PelAddr += 50;
						PelPtr1 += (Pitch >> 1) - 512;
						PelPtr2 += (Pitch >> 1) - 512;
					}
				} break;
			}
		}

		if(over)
		{
			SDL_UnlockYUVOverlay(over);
			SDL_Rect target;
			target.x = 0;
			target.w = front->w;
			target.y = 0;
			target.h = front->h;
			SDL_DisplayYUVOverlay(over, &target);
		}
		else
		{
			SDL_UnlockSurface(front);
			SDL_UpdateRect(front, 0, 0, 0, 0);
		}
	}

	while(BReadPtr != BWritePtr)
	{
		BColour = BorderQueue[BReadPtr].Colour;
		BReadPtr = (BReadPtr+1)&1023;
	}
	Time += cycles;

	return 0;
}

void CScreen::SetupGraphics()
{
	delete[] Lookup; Lookup = NULL;
	delete[] MultiLookup; MultiLookup = NULL;
	if(over) {SDL_FreeYUVOverlay(over); over = NULL;}
	if(front) SDL_FreeSurface(front);

	/* try for an overlay surface first if fullscreen - force to 8bpp colour as video memory may be a concern here*/
	if(AllowOverlay && (GFXFlags & SDL_FULLSCREEN))
	{
		while(1)
		{
			front = SDL_SetVideoMode(1024, 768, 8, GFXFlags);
			if(front) break;
			GFXFlags ^= SDL_FULLSCREEN;
		}

		if(GFXFlags&SDL_FULLSCREEN)
		{
			Uint32 OverlayTypes[] =
			{
				SDL_YUY2_OVERLAY,
				SDL_YVYU_OVERLAY,
				SDL_UYVY_OVERLAY,
				0
			};
			int OverlayPtr = 0;

			while(!over && OverlayTypes[OverlayPtr])
			{
				over = SDL_CreateYUVOverlay(640, 480, OverlayTypes[OverlayPtr], front);

				if(over->hw_overlay != 1)
				{
					SDL_FreeYUVOverlay(over);
					OverlayPtr++;
					over = NULL;
				}
			}
		}
	}

	if(over)
	{
		Uint8	YTable[16];
		Uint8	UTable[16];
		Uint8	VTable[16];

		/* initialise Y, U, V tables */
		int c = 16;
		while(c--)
		{
			float R = (float)RTable[c] / 255.0f;
			float G = (float)GTable[c] / 255.0f;
			float B = (float)BTable[c] / 255.0f;

			float Y = 16	+ R*65.481f	+ G*128.553f	+ B*24.966f;
			float U = 128	- R*37.797f - G*74.203f		+ B*112.0f;
			float V = 128	+ R*112.0f	- G*93.786f		- B*18.214f;

			YTable[c] = (int)Y;
			if(Y < 0) YTable[c] = 0;
			if(Y > 255) YTable[c] = 255;

			UTable[c] = (int)U;
			if(U < 0) UTable[c] = 0;
			if(U > 255) UTable[c] = 255;

			VTable[c] = (int)V;
			if(V < 0) VTable[c] = 0;
			if(V > 255) VTable[c] = 255;
		}

		int YOffs1, YOffs2, UOffs, VOffs;
		switch(over->format)
		{
			default:
			case SDL_YVYU_OVERLAY: YOffs1 = 0; YOffs2 = 2; VOffs = 1; UOffs = 3; break;
			case SDL_YUY2_OVERLAY: YOffs1 = 0; YOffs2 = 2; VOffs = 3; UOffs = 1; break;
			case SDL_UYVY_OVERLAY: YOffs1 = 1; YOffs2 = 3; UOffs = 0; VOffs = 2; break;
		}

		Lookup = new Uint8[(1 << 15)*2*8];
		MultiLookup = new Uint8[4 << 15];
		c = 1 << 15;
		while(c--)
		{
			Uint32 Col1, Col2;
			Col1 = (c >> 8)&7;
			Col2 = (c >> 11)&15;
			Col1 |= Col2&8;

			int sc = c&255;
			int scc = 4;
			int byteno = c << 4;
			while(scc--)
			{
				Uint32 Colour1 = (sc&0x80) ? Col1 : Col2, Colour2 = (sc&0x40) ? Col1 : Col2;

				Lookup[byteno + YOffs1] = YTable[Colour1];
				Lookup[byteno + YOffs2] = YTable[Colour2];

				Lookup[byteno + VOffs] = (VTable[Colour1] + VTable[Colour2]) >> 1;
				Lookup[byteno + UOffs] = (UTable[Colour1] + UTable[Colour2]) >> 1;

				byteno += 4;
				sc <<= 2;
			}
		}
	}
	else
	{
		/* 8bpp mode will be faster, so if full screen try for it (unless multiplexed - then aim for 16) */
		int tries = 2;
		while(tries--)
		{
			front = SDL_SetVideoMode(640, 480, (GFXFlags&SDL_FULLSCREEN) ? (Multi ? 16 : 8) : 0, GFXFlags);
			if(front) break;
			GFXFlags ^= SDL_FULLSCREEN;
		}
		if(tries < 0)
			return;

		/* initalise relevant RGB table */
		int c = 16;
		Uint32	RGBTable32[16];
		Uint16	RGBTable16[16];
		Uint8	RGBTable8[16];
		
		/* pragmatic decision: Lookup not to be larger than 64kb. Cache is more important! */
		Lookup = new Uint8[front->format->BytesPerPixel * (1 << 14)];
		MultiLookup = new Uint8[front->format->BytesPerPixel * (1 << 15)];
		switch(front->format->BytesPerPixel)
		{
			case 1:{
				SDL_Color SpecPal[16] =
				{
					{0, 0, 0}, {0, 0, 203}, {203, 0, 0}, {203, 0, 203}, {0, 203, 0}, {0, 203, 203}, {203, 203, 0}, {203, 203, 203},
					{0, 0, 0}, {0, 0, 255}, {255, 0, 0}, {255, 0, 255}, {0, 255, 0}, {0, 255, 255}, {255, 255, 0}, {255, 255, 255},
				};
				SDL_SetColors(front, SpecPal, 16, 16);
				
				while(c--)
					RGBTable8[c] = SDL_MapRGB(front->format, RTable[c], GTable[c], BTable[c]);

				c = 1 << 11;
				while(c--)
				{
					Uint32 Col1, Col2;
					Col1 = (c >> 4)&7;
					Col2 = (c >> 7)&15;
					Col1 |= Col2&8;

					Col1 = RGBTable8[Col1];
					Col2 = RGBTable8[Col2];

					int sc = c&15;
					int scc = 4;
					int byteno = c*8*front->format->BytesPerPixel;
					while(scc--)
					{
						Uint32 Colour = (sc&0x8) ? Col1 : Col2;

						Lookup[byteno] = Lookup[byteno+1] = Colour;
						byteno += 2;

						sc <<= 1;
					}
				}
			}break;

			case 2:
				while(c--)
					RGBTable16[c] = SDL_MapRGB(front->format, RTable[c], GTable[c], BTable[c]);

				c = 1 << 11;
				while(c--)
				{
					Uint16 Col1, Col2;
					Col1 = (c >> 4)&7;
					Col2 = (c >> 7)&15;
					Col1 |= Col2&8;

					Col1 = RGBTable16[Col1];
					Col2 = RGBTable16[Col2];

					int sc = c&15;
					int scc = 4;
					int byteno = c*8*front->format->BytesPerPixel;
					while(scc--)
					{
						Uint16 Colour = (sc&0x8) ? Col1 : Col2;

						Lookup[byteno+0] = Lookup[byteno+2] = ((Uint8 *)&Colour)[0];
						Lookup[byteno+1] = Lookup[byteno+3] = ((Uint8 *)&Colour)[1];
						byteno += front->format->BytesPerPixel*2;

						sc <<= 1;
					}
				}
			break;

			case 3:
			case 4:{
				while(c--)
					RGBTable32[c] = SDL_MapRGB(front->format, RTable[c], GTable[c], BTable[c]);

				c = 1 << 11;
				while(c--)
				{
					Uint32 Col1, Col2;
					Col1 = (c >> 4)&7;
					Col2 = (c >> 7)&15;
					Col1 |= Col2&8;

					Col1 = RGBTable32[Col1];
					Col2 = RGBTable32[Col2];

					int sc = c&15;
					int scc = 4;
					int byteno = c*8*front->format->BytesPerPixel;
					while(scc--)
					{
						Uint32 Colour = (sc&0x08) ? Col1 : Col2;
						
						Lookup[byteno+0] = Lookup[byteno+front->format->BytesPerPixel+0] = ((Uint8 *)&Colour)[0];
						Lookup[byteno+1] = Lookup[byteno+front->format->BytesPerPixel+1] = ((Uint8 *)&Colour)[1];
						Lookup[byteno+2] = Lookup[byteno+front->format->BytesPerPixel+2] = ((Uint8 *)&Colour)[2];
						if(front->format->BytesPerPixel == 4)
							Lookup[byteno+3] = Lookup[byteno+front->format->BytesPerPixel+3] = ((Uint8 *)&Colour)[3];

						byteno += front->format->BytesPerPixel*2;

						sc <<= 1;
					}
				}
			}break;
		}
	}

	SetMultiPalette(MultiPalette);

	if(GFXFlags&SDL_FULLSCREEN) SDL_ShowCursor(SDL_DISABLE);
	SDL_WM_SetCaption("Dan Dare: Pilot of the Future", "Dan Dare: Pilot of the Future");
}

const Uint8 *CScreen::GetMultiPalette()
{
	return MultiPalette;
}

void CScreen::SetMultiPalette(Uint8 *Table)
{
	memcpy(MultiPalette, Table, 129*256*3);

	/* if in paletted mode, try to apply supplied compromise palette */
	if(!over && (front->format->BytesPerPixel == 1))
	{
		SDL_Color Pal[256];

		int c = 256;
		while(c--)
		{
			Pal[c].r = Table[ 128*256*3 + c*3 + 0];
			Pal[c].g = Table[ 128*256*3 + c*3 + 1];
			Pal[c].b = Table[ 128*256*3 + c*3 + 2];
		}

		SDL_SetColors(front, Pal, 0, 256);
	}

	/* table is indexed with 7 bit attribute in high part, index in low part to produce R, G, B triplet */
	int attr = 128;
	while(attr--)
	{
		int index = 256;
		while(index--)
		{
			int tabindex = ((attr << 8) | index) * 3;

			if(over)
			{
				float R = (float)Table[tabindex + 0] / 255.0f;
				float G = (float)Table[tabindex + 1] / 255.0f;
				float B = (float)Table[tabindex + 2] / 255.0f;

				float Y = 16	+ R*65.481f	+ G*128.553f	+ B*24.966f;
				float U = 128	- R*37.797f - G*74.203f		+ B*112.0f;
				float V = 128	+ R*112.0f	- G*93.786f		- B*18.214f;

				if(Y < 0) Y = 0;
				if(Y > 255) Y = 255;

				if(U < 0) U = 0;
				if(U > 255) U = 255;

				if(V < 0) V = 0;
				if(V > 255) V = 255;

				int dstindex = (index | (attr << 8)) << 2;

				MultiLookup[dstindex + 0] = (int)Y;
				MultiLookup[dstindex + 1] = (int)U;
				MultiLookup[dstindex + 2] = (int)V;
			}
			else
			{
				switch(front->format->BytesPerPixel)
				{
					case 1:{
						Uint8 Colour = SDL_MapRGB(front->format, Table[tabindex + 0], Table[tabindex + 1], Table[tabindex + 2]);

						int dstindex = (index | (attr << 8))*front->format->BytesPerPixel;
						MultiLookup[dstindex] = Colour;
					} break;
					case 2:{
						Uint16 Colour = SDL_MapRGB(front->format, Table[tabindex + 0], Table[tabindex + 1], Table[tabindex + 2]);

						int dstindex = (index | (attr << 8))*front->format->BytesPerPixel;
						MultiLookup[dstindex] = ((Uint8 *)&Colour)[0];
						MultiLookup[dstindex+1] = ((Uint8 *)&Colour)[1];
					} break;
					case 3:{
						Uint32 Colour = SDL_MapRGB(front->format, Table[tabindex + 0], Table[tabindex + 1], Table[tabindex + 2]);

						int dstindex = (index | (attr << 8))*front->format->BytesPerPixel;
						MultiLookup[dstindex] = ((Uint8 *)&Colour)[0];
						MultiLookup[dstindex+1] = ((Uint8 *)&Colour)[1];
						MultiLookup[dstindex+2] = ((Uint8 *)&Colour)[2];
					} break;
					case 4:{
						Uint32 Colour = SDL_MapRGB(front->format, Table[tabindex + 0], Table[tabindex + 1], Table[tabindex + 2]);

						int dstindex = (index | (attr << 8))*front->format->BytesPerPixel;
						MultiLookup[dstindex] = ((Uint8 *)&Colour)[0];
						MultiLookup[dstindex+1] = ((Uint8 *)&Colour)[1];
						MultiLookup[dstindex+2] = ((Uint8 *)&Colour)[2];
						MultiLookup[dstindex+3] = ((Uint8 *)&Colour)[3];
					} break;
				}
			}
		}
	}
}


CScreen::CScreen()
{
	BWritePtr = BReadPtr = Time = BColour = 0;
	Lookup = NULL;
	MultiLookup = NULL;
	over = NULL;
	front = NULL;
	GFXFlags = SDL_HWSURFACE;
	Multi = false;
	BehindNo = 0;
	AllowOverlay = false;
	FlagsDirty = false;

	FILE *Options = Resourcefopen("options.dat", "rb");
	if(Options)
	{
		Uint8 Flags = fgetc(Options);
		AllowOverlay = (Flags&1) ? true : false;
		if(Flags&2) GFXFlags |= SDL_FULLSCREEN;
		fclose(Options);
	}

	/* set default palette for multiplexing */
	int attr = 128;
	while(attr--)
	{
		MultiPalette[((attr << 8) * 3) + 0] = RTable[attr >> 3];
		MultiPalette[((attr << 8) * 3) + 1] = GTable[attr >> 3];
		MultiPalette[((attr << 8) * 3) + 2] = BTable[attr >> 3];

		MultiPalette[(((attr << 8) | 0xff)) * 3 + 0] = RTable[(attr&7) | ((attr >> 3)&0x8)];
		MultiPalette[(((attr << 8) | 0xff)) * 3 + 1] = GTable[(attr&7) | ((attr >> 3)&0x8)];
		MultiPalette[(((attr << 8) | 0xff)) * 3 + 2] = BTable[(attr&7) | ((attr >> 3)&0x8)];
	}
	attr = 16;
	while(attr--)
	{
		MultiPalette[ 128*256*3 + 48 + attr*3 + 0] = RTable[attr];
		MultiPalette[ 128*256*3 + 48 + attr*3 + 1] = GTable[attr];
		MultiPalette[ 128*256*3 + 48 + attr*3 + 2] = BTable[attr];
	}


	SetupGraphics();
}

void CScreen::ToggleFullscreen()
{
	GFXFlags ^= SDL_FULLSCREEN;
	SetupGraphics();
}

CScreen::~CScreen()
{
	FILE *Options = Resourcefopen("options.dat", "wb");
	if(Options)
	{
		fputc(
			(AllowOverlay ? 1 : 0) |
			((GFXFlags & SDL_FULLSCREEN) ? 2 : 0),
			Options);
		fclose(Options);
	}

	if(over)	SDL_FreeYUVOverlay(over);
	if(front)	SDL_FreeSurface(front);
	delete[] Lookup;
	delete[] MultiLookup;
}

void CScreen::Attach(CZ80 *CPU)
{
	/* set default graphics page */
	CPU->SetCopyMask((2+5) << 14);

	/* setup timing */
	CPU->SetTiming(0, false);
	CPU->SetTiming(1, false);
	CPU->SetTiming(10, false);

	int sc = 8;
	while(sc--)
	{
		CPU->SetTiming(sc+2, (sc&1) ? true : false);
	}

	/* fill contended timings for frame */
	Uint8 *ContTab = CPU->GetContendedTable();

	int y = 192;
	int offset = CONTENDED_START;
	while(y--)
	{
		int x = 16;
		while(x--)
		{
			ContTab[offset] = 6;
			ContTab[offset+1] = 5;
			ContTab[offset+2] = 4;
			ContTab[offset+3] = 3;
			ContTab[offset+4] = 2;
			ContTab[offset+5] = 1;

			offset += 8;
		}

		offset += 100;
	}

	int LineAddr = 0;
	memset(CopySources, 0, sizeof(int)*35454);

	offset = CONTENDED_START >> 1;
	y = 192;
	while(y--)
	{
		int x = 32;
		while(x--)
		{
 			CopySources[offset] = LineAddr & 0x181f;
 			CopySources[offset] |= (LineAddr & 0x0e0) << 3;
 			CopySources[offset] |= (LineAddr & 0x700) >> 3;

			CopySources[offset+1] = LineAddr & 0x1f;
			CopySources[offset+1] |= (LineAddr & 0x1f00) >> 3;
			CopySources[offset+1] += 6144;

			LineAddr++;
			offset += 2;
		}

		offset += 50;
	}

	CPU->SetCopyAddrs(CopySources);
	CPU->SetCopyTarget(PelData8, PelData64);
	CPU->SetGatherTimes(CONTENDED_START, 128, 228, 192);
}

void CScreen::SetFlags(Uint32 Flags)
{
	bool OMulti = Multi, OAO = AllowOverlay;

	Multi = (Flags&CSF_MULTIPLEXED) ? true : false;
	if(GFXFlags & SDL_FULLSCREEN) AllowOverlay = (Flags&CSF_ALLOWOVERLAY) ? true : false;

	FlagsDirty = (Multi != OMulti) || (AllowOverlay != OAO);
}

Uint32 CScreen::GetFlags()
{
	return (Multi ? CSF_MULTIPLEXED : 0) | (AllowOverlay ? CSF_ALLOWOVERLAY : 0);
}

bool CScreen::HasScreen()
{
	return front ? true : false;
}
