#include "Tape.h"
#include <stdio.h>
#include "SDL.h"
#include "z80.h"
#include "ULA.h"

CTape::CTape(CULA *u)
{
	Tape = NULL;
	TrapFunction = NULL;
	ULA = u;
}

CTape::~CTape()
{
	Close();
}

bool CTape::Open(char *name, unsigned int Timestamp)
{
	Tape = fopen(name, "rb");
	return Tape ? true : false;
}

void CTape::Close()
{
	if(Tape)
	{
		fclose(Tape);
		Tape = NULL;
	}
}

bool CTape::ReadBit(unsigned int Timestamp)
{
	return false;
}

void CTape::TapeTrap(void *ptr, CZ80 &CPU, Uint16 Addr)
{
	((CTape *)ptr)->TapeTrap(CPU, Addr);
}

void CTape::TapeTrap(CZ80 &CPU, Uint16 Addr)
{
	ULA->SetEnter(false);
	static char FName[11];

	int Read, Write;
	CPU.GetPaged(0, Read, Write);
	if(!Tape || (Read != 1)) return;

	long spos = ftell(Tape);
	RegisterState rs;

	rs = CPU.GetState();

	// read block length
	Uint16 Length;
	Length = fgetc(Tape);
	Length |= fgetc(Tape) << 8;

	if(feof(Tape))
	{
		spos = 0;
		fseek(Tape, 0, SEEK_SET);
		Length = fgetc(Tape);
		Length |= fgetc(Tape) << 8;
	}

//	printf("block length:%d DE:%d", Length, rs.DE.a);

	if(Length < rs.DE.a)
	{
		rs.AF.b.b0 &= ~1;
		fseek(Tape, spos, SEEK_SET);
//		printf("\n");

		CPU.SetState(rs);
		return;
	}

	Uint8 A = fgetc(Tape); Length--;
//	printf(", a':%02x v %02x", rs.AF2.b.b1, A);
	if(A != rs.AF2.b.b1)
	{
		rs.AF.b.b0 &= ~1;
		fseek(Tape, spos, SEEK_SET);
//		printf("\n");

		CPU.SetState(rs);
		return;
	}

//	printf(", carry':%d", rs.AF2.b.b0&1);
	if(rs.AF2.b.b0&1)
	{
		/* write remaining data to IX+ (and duplicate file name) */
		Uint16 Addr, C;

		FName[10] = '\0';

		C = rs.DE.a;
		Addr = rs.IX.a;
		while(C--)
		{
			Uint8 T8 = fgetc(Tape); Length--;
			CPU.WriteMemory(Addr, T8);

			if(((Addr-rs.IX.a) < 11) && ((Addr-rs.IX.a) > 0))
				FName[Addr-rs.IX.a-1] = T8;

			Addr++;
		}

		rs.AF.b.b0 |= 1;
//		printf(", loading to ix:%04x [%s]\n", rs.IX.a, FName);
	}
	else
	{
//		printf("carry' clear: verifying\n", rs.IX.a);
		rs.AF.b.b0 &= ~1;
	}

	fseek(Tape, Length, SEEK_CUR);
	rs.PC.a = 0x05e2;
	CPU.SetState(rs);

	/*printf("reflection:\n");
	for(int addr = 0x6400; addr < 0x6500; addr++)
	{
		Uint8 iv;
		CPU.ReadMemory(addr, iv);
		printf("%02x\n", iv);
	}*/
	if(TrapFunction)
		TrapFunction(TrapParam, CPU, rs.IX.a, rs.DE.a);
	return;
}

void CTape::SetTrapFunction( void (* Trap)(void *, CZ80 &, Uint16, Uint16), void *param )
{
	TrapParam = param;
	TrapFunction = Trap;
}
