#include "z80.h"
#include "SDL.h"
#include "SDL_thread.h"
#include <memory.h>

Uint8 CZ80::ParityTab[256];

#define ReadFromAF(r)\
	r.b.b1 = AReg;\
	r.b.b0 = (Misc&0x28) | Carry | (LastOp ? 2 : 0) | ((Overflow >> 5)&0x4) | HalfCarry | (Zero ? 0 : 0x40) | (Sign&0x80);

#define WriteToAF(r)\
	AReg = r.b.b1;\
	Misc = Sign = r.b.b0;\
	Zero = (r.b.b0&0x40)^0x40;\
	HalfCarry = r.b.b0&0x10;\
	Carry = r.b.b0&1;\
	LastOp = (r.b.b0&2) ? true : false;\
	Overflow = (r.b.b0&0x4) << 5;
	
void CZ80::DoBreak()
{
	TotalCycleCount += CyclesToRun;
	SubCycleCount -= CyclesToRun;
	CopyTime = SubCycleCount;
	SDL_SemPost(StopSemaphore);
	SDL_SemWait(GoSemaphore);
}

#define Break()	\
	SDL_SemPost(StopSemaphore);\
	SDL_SemWait(GoSemaphore);

#define DoCycles(n)\
	SubCycleCount += n;\
	while(CopyTime < SubCycleCount)\
	{\
		TargetBuffer8[CopyTime >>1] = MemoryPool8[ CopySource[CopyTime >> 1] | CopySourceMask ];\
		TargetBuffer64[CopyTime >>1] = MemoryPool64[ CopySource[CopyTime >> 1] | CopySourceMask ];\
		CopyTime += 2;\
	}\
	if(SubCycleCount >= CyclesToRun)\
		DoBreak();
	
/*	{\
		TotalCycleCount += CyclesToRun;\
		SubCycleCount -= CyclesToRun;\
		CopyTime = SubCycleCount;\
		Break();\
	}*/

#define Read8(n, v)		MemAccess(n.a); v = Read8Ptrs[(n.a) >> 14][(n.a)&16383]; DoCycles(3);
#define Write8(n, v)	MemAccess(n.a); Write8Ptrs[(n.a) >> 14][(n.a)&16383] = v; DoCycles(3);

#define Read64(n, v, rv)	MemAccess(n.a); v = Read8Ptrs[(n.a) >> 14][(n.a)&16383]; rv = Read64Ptrs[(n.a) >> 14][(n.a)&16383]; DoCycles(3);
#define Write64(n, v, rv)	MemAccess(n.a); Write8Ptrs[(n.a) >> 14][(n.a)&16383] = v; Write64Ptrs[(n.a) >> 14][(n.a)&16383] = rv; DoCycles(3);

#define MemAccess(n)	DoCycles(TimePtrs[n >> 14][SubCycleCount]);
#define IOAccess(n)		if(IOContention[n >> 5] & (1 << (n&31))) {DoCycles(TimePool[70000 + SubCycleCount]);}

#define Push16(r)\
	SP.a--;\
	Write8(SP, r.b.b1);\
	SP.a--;\
	Write8(SP, r.b.b0);

#define WPush16(r, rvh, rvl)\
	SP.a--;\
	Write64(SP, r.b.b1, rvh);\
	SP.a--;\
	Write64(SP, r.b.b0, rvl);

#define Push16Instr(r)				DoCycles(1); Push16(r);
#define WPush16Instr(r, rvh, rvl)	DoCycles(1); WPush16(r, rvh, rvl);

#define Pop16(r)\
	Read8(SP, r.b.b0);\
	SP.a++;\
	Read8(SP, r.b.b1);\
	SP.a++;

#define WPop16(r, rvh, rvl)\
	Read64(SP, r.b.b0, rvl);\
	SP.a++;\
	Read64(SP, r.b.b1, rvh);\
	SP.a++;

#define Dec16(r)\
	r.a--;\
	DoCycles(2);

#define Inc16(r)\
	r.a++;\
	DoCycles(2);

#define Or(r)\
	AReg |= r;\
	Sign = Zero = Misc = AReg; HalfCarry = Carry = 0; LastOp = false;\
	Overflow = ParityTab[AReg];

#define Xor(r)\
	AReg ^= r;\
	Sign = Zero = Misc = AReg; HalfCarry = Carry = 0; LastOp = false;\
	Overflow = ParityTab[AReg];

#define And(r)\
	AReg &= r;\
	Sign = Zero = Misc = AReg; HalfCarry = 0x10; Carry = 0; LastOp = false;\
	Overflow = ParityTab[AReg];

#define WOr(r, rv)\
	Or(r);\
	CarryBig = 0;\
	ABig |= rv
/*	ABig =		((rv&0x00000000000000ff) ? (rv&0x00000000000000ff) : (ABig&0x00000000000000ff)) |\
				((rv&0x000000000000ff00) ? (rv&0x000000000000ff00) : (ABig&0x000000000000ff00)) |\
				((rv&0x0000000000ff0000) ? (rv&0x0000000000ff0000) : (ABig&0x0000000000ff0000)) |\
				((rv&0x00000000ff000000) ? (rv&0x00000000ff000000) : (ABig&0x00000000ff000000)) |\
				((rv&0x000000ff00000000) ? (rv&0x000000ff00000000) : (ABig&0x000000ff00000000)) |\
				((rv&0x0000ff0000000000) ? (rv&0x0000ff0000000000) : (ABig&0x0000ff0000000000)) |\
				((rv&0x00ff000000000000) ? (rv&0x00ff000000000000) : (ABig&0x00ff000000000000)) |\
				((rv&0xff00000000000000) ? (rv&0xff00000000000000) : (ABig&0xff00000000000000));*/

#define WXor(r, rv)\
	Xor(r);\
	ABig ^= rv;\
	CarryBig = 0;

#define WAnd(r, rv)\
	And(r);\
	ABig &= rv;\
	CarryBig = 0;

#define Bit(r, v)\
	Sign = Misc = Zero = (r)&(v);\
	Overflow = Zero ? 0 : 0x80;\
	LastOp = false;\
	HalfCarry = 0x10;

#define JR(cond)\
	Read8(PC, TempValue);\
	if(cond)\
	{\
		MemAccess(PC.a);\
		DoCycles(1);\
		MemAccess(PC.a);\
		DoCycles(1);\
		MemAccess(PC.a);\
		DoCycles(1);\
		MemAccess(PC.a);\
		DoCycles(1);\
		MemAccess(PC.a);\
		DoCycles(1);\
		PC.a += (Sint8)TempValue;\
	}\
	PC.a++;

#define Ret(cond)\
	DoCycles(1);\
	if(cond) {Pop16(PC);}

#define JP(cond)\
	Load16nn(Addr);\
	if(cond)	PC.a = Addr.a;

#define Call(cond)	\
	Load16nn(Addr);\
	MemAccess(PC.a);\
	DoCycles(1);\
	if(cond)\
	{\
		SP.a--;\
		Write8(SP, PC.b.b1);\
		SP.a--;\
		Write8(SP, PC.b.b0);\
		PC.a = Addr.a;\
	}

#define Load1616(r1, r2)\
	r1.a = r2.a;\
	DoCycles(2);

#define Load8n(r)\
	Read8(PC, r);\
	PC.a++;

#define WLoad8n(r, rv)\
	Read64(PC, r, rv);\
	PC.a++;

#define Load16nn(r)\
	Read8(PC, r.b.b0); PC.a++;\
	Read8(PC, r.b.b1); PC.a++;

#define WLoad16nn(r, rvh, rvl)\
	Read64(PC, r.b.b0, rvl); PC.a++;\
	Read64(PC, r.b.b1, rvh); PC.a++;

#define Load16nni(r)\
	Load16nn(Addr);\
	Read8(Addr, r.b.b0); Addr.a++;\
	Read8(Addr, r.b.b1);

#define WLoad16nni(r, rvh, rvl)\
	Load16nn(Addr);\
	Read64(Addr, r.b.b0, rvl); Addr.a++;\
	Read64(Addr, r.b.b1, rvh);

#define Load8nn(r)\
	Load16nn(Addr);\
	Read8(Addr, r);

#define WLoad8nn(r, rv)\
	Load16nn(Addr);\
	Read64(Addr, r, rv);

#define Loadnn16(r)\
	Load16nn(Addr);\
	Write8(Addr, r.b.b0); Addr.a++;\
	Write8(Addr, r.b.b1);

#define WLoadnn16(r, rvh, rvl)\
	Load16nn(Addr);\
	Write64(Addr, r.b.b0, rvl); Addr.a++;\
	Write64(Addr, r.b.b1, rvh);

#define Inc8(r)\
	HalfCarry = ((r&0xf)+1)&0x10;\
	Overflow = (r == 127) ? 0x80 : 0;\
	r++;\
	LastOp = false; Sign = Misc = Zero = r;

#define Dec8(r)\
	HalfCarry = ((r&0xf)-1)&0x10;\
	Overflow = (r == 128) ? 0x80 : 0;\
	r--;\
	LastOp = true; Sign = Misc = Zero = r;

#define StHL()	Write8(HL, TempValue);
#define WStHL()	Write64(HL, TempValue, TempValueBig);

/*
	add - overflow occurs if two numbers of the same sign added yield a number with a
	different sign

	~AReg^n = 1 if numbers of same sign
	AReg^Temp16.b.b0 = 1 if result is of different sign

	sub - overflow occurs if the opposite is true

*/

#define Add16(r, n)\
	LastOp = false;\
	Temp32 = r.a + n.a;\
	HalfCarry = (((r.a&0xfff) + (n.a&0xfff)) >> 8)&0x10;\
	Misc = Temp32 >> 8;\
	Carry = Temp32 >> 16;\
	r.a = Temp32;\
	DoCycles(7);

#define Adc16(n)\
	LastOp = false;\
	Temp32 = HL.a + n.a + Carry;\
	HalfCarry = (((HL.a&0xfff) + (n.a&0xfff) + Carry) >> 8)&0x10;\
	Sign = Misc = Temp32 >> 8;\
	Zero = Temp32 | Sign;\
	Carry = Temp32 >> 16;\
	Overflow = (~(HL.a^n.a)&(HL.a^Temp32)) >> 8;\
	HL.a = Temp32;\
	DoCycles(7);

#define Sbc16(n)\
	LastOp = true;\
	Temp32 = HL.a - n.a - Carry;\
	HalfCarry = (((HL.a&0xfff) - (n.a&0xfff) - Carry) >> 8)&0x10;\
	Misc = Sign = Temp32 >> 8;\
	Zero = Temp32 | Sign;\
	Carry = (Temp32 >> 16)&1;\
	Overflow = ((HL.a^n.a)&(HL.a^Temp32)) >> 8;\
	HL.a = Temp32;\
	DoCycles(7);

#define Add8(n)\
	LastOp = false;\
	Temp16.a = AReg + n;\
	HalfCarry = ((AReg&0xf) + (n&0xf))&0x10;\
	Carry = Temp16.b.b1;\
	Overflow = ~(AReg^n)&(AReg^Temp16.b.b0);\
	Misc = Sign = Zero = AReg = Temp16.b.b0;

#define Adc8(n)\
	LastOp = false;\
	Temp16.a = AReg + n + Carry;\
	HalfCarry = ((AReg&0xf) + (n&0xf) + Carry)&0x10;\
	Carry = Temp16.b.b1;\
	Overflow = ~(AReg^n)&(AReg^Temp16.b.b0);\
	Misc = Sign = Zero = AReg = Temp16.b.b0;

#define Sbc8(n)\
	LastOp = true;\
	Temp16.a = AReg - n - Carry;\
	HalfCarry = ((AReg&0xf) - (n&0xf) - Carry)&0x10;\
	Carry = Temp16.b.b1&1;\
	Overflow = (AReg^n)&(AReg^Temp16.b.b0);\
	Misc = Sign = Zero = AReg = Temp16.b.b0;

#define Sub8(n)\
	LastOp = true;\
	Temp16.a = AReg - n;\
	HalfCarry = ((AReg&0xf) - (n&0xf))&0x10;\
	Carry = Temp16.b.b1&1;\
	Overflow = (AReg^n)&(AReg^Temp16.b.b0);\
	Misc = Sign = Zero = AReg = Temp16.b.b0;

#define Cp(n)\
	Temp32 = AReg - n;\
	Misc = n;\
	HalfCarry = ((AReg&0x0f) - (n&0x0f))&0x10;\
	Sign = Zero = Temp32;\
	Carry = (Temp32 >> 8)&1;\
	Overflow = (AReg^n)&(AReg^Temp32);\
	LastOp = true;

#define Srl(n)\
	LastOp = false; HalfCarry = 0;\
	Carry = n&1;\
	n >>= 1;\
	Overflow = ParityTab[Misc = Sign = Zero = n];

#define Sra(n)\
	LastOp = false; HalfCarry = 0;\
	Carry = n&1;\
	n = (n >>1) | (n&0x80);\
	Overflow = ParityTab[Misc = Sign = Zero = n];

#define Rr(n)\
	LastOp = false; HalfCarry = 0;\
	Temp8 = Carry;\
	Carry = n&1;\
	n = (n >> 1) | (Temp8 << 7);\
	Overflow = ParityTab[Misc = Sign = Zero = n];

#define Rrc(n)\
	LastOp = false; HalfCarry = 0;\
	Carry = n&1;\
	n = (n >> 1) | (Carry << 7);\
	Overflow = ParityTab[Misc = Sign = Zero = n];

#define Sll(n)\
	LastOp = false; HalfCarry = 0;\
	Carry = n >> 7;\
	n = (n << 1) | 1;\
	Overflow = ParityTab[Misc = Sign = Zero = n];

#define Sla(n)\
	LastOp = false; HalfCarry = 0;\
	Carry = n >> 7;\
	n <<= 1;\
	Overflow = ParityTab[Misc = Sign = Zero = n];

#define Rl(n)\
	LastOp = false; HalfCarry = 0;\
	Temp8 = Carry;\
	Carry = n >> 7;\
	n = (n << 1) | Temp8;\
	Overflow = ParityTab[Misc = Sign = Zero = n];

#define Rlc(n)\
	LastOp = false; HalfCarry = 0;\
	Carry = n >> 7;\
	n = (n << 1) | Carry;\
	Overflow = ParityTab[Misc = Sign = Zero = n];


#define WSrl(n, nr)\
	Srl(n);\
	CarryBig = nr&0xff;\
	nr >>= 8;

#define WSra(n, nr)\
	Sra(n);\
	CarryBig = nr&0xff;\
	nr = (nr >> 8) | (nr&0xff00000000000000ull);

#define WRr(n, nr)\
	Rr(n);\
	Temp64 = CarryBig;\
	CarryBig = nr&0xff;\
	nr = (nr >> 8) | (Temp64 << 56);

#define WRrc(n, nr)\
	Rrc(n);\
	CarryBig = nr&0xff;\
	nr = (nr >> 8) | (CarryBig << 56);

#define WSll(n, nr)\
	Sll(n);\
	CarryBig = nr >> 56;\
	nr = (nr << 8) | 0xff;

#define WSla(n, nr)\
	Sla(n);\
	CarryBig = nr >> 56;\
	nr <<= 8;

#define WRl(n, nr)\
	Rl(n);\
	Temp64 = CarryBig;\
	CarryBig = nr >> 56;\
	nr = (nr << 8) | Temp64;

#define WRlc(n, nr)\
	Rlc(n);\
	CarryBig = nr >> 56;\
	nr = (n << 8) | CarryBig;

#define Res(r, v, rv, vr)	(r) &= ~(v); (rv) &= ~(vr);
#define Set(r, v, rv, vr)	(r) |= (v); (rv) |= (vr);

#define Out(port, value)\
	DoCycles(1);\
	IOAccess(port);\
	DoCycles(3);\
	WritePort(port, value, TotalCycleCount+SubCycleCount);

#define In(port, value)\
	DoCycles(1);\
	IOAccess(port);\
	DoCycles(3);\
	ReadPort(port, value, TotalCycleCount+SubCycleCount);\
	Overflow = ParityTab[Sign = Misc = Zero = value];\
	HalfCarry = 0;\
	LastOp = false;

#define CPICompare()\
	Read8(HL, TempValue);\
\
	Temp16.a = AReg - TempValue;\
	HalfCarry = ((AReg&0x0f) - (TempValue&0x0f))&0x10;\
	Sign = Zero = Temp16.b.b0;\
	LastOp = true;\
	if(HalfCarry) Temp16.a--;\
	Misc = ((Temp16.b.b0&0x01) << 4) | (Temp16.b.b0&0x08);\
\
	BC.a--;\
	Overflow = BC.a ? 0x80 : 0;

#define CopyHLToDE()\
	Read64(HL, TempValue, TempValueBig);\
	Write64(DE, TempValue, TempValueBig);\
\
	TempValue += AReg;\
	Misc = TempValue&0x8 | ((TempValue&0x2) << 4);\
	LastOp = false;\
	HalfCarry = 0;\
\
	BC.a--;\
	Overflow = BC.a ? 0x80 : 0;

#define SetGroup(base, value, vr)\
	case base:		Set(BReg, value, BBig, vr);	break;\
	case base+1:	Set(CReg, value, CBig, vr);	break;\
	case base+2:	Set(DReg, value, DBig, vr)	break;\
	case base+3:	Set(EReg, value, EBig, vr);	break;\
	case base+4:	Set(HReg, value, HBig, vr);	break;\
	case base+5:	Set(LReg, value, LBig, vr);	break;\
	case base+6:	Read64(HL, TempValue, TempValueBig); Set(TempValue, value, TempValueBig, vr); WStHL(); break;\
	case base+7:	Set(AReg, value, ABig, vr);	break;

#define ResGroup(base, value, vr)\
	case base:		Res(BReg, value, BBig, vr);	break;\
	case base+1:	Res(CReg, value, CBig, vr);	break;\
	case base+2:	Res(DReg, value, DBig, vr);	break;\
	case base+3:	Res(EReg, value, EBig, vr);	break;\
	case base+4:	Res(HReg, value, HBig, vr);	break;\
	case base+5:	Res(LReg, value, LBig, vr);	break;\
	case base+6:	Read64(HL, TempValue, TempValueBig); Res(TempValue, value, TempValueBig, vr); WStHL(); break;\
	case base+7:	Res(AReg, value, ABig, vr);	break;

#define BitGroup(base, value, vr)\
	case base:		Bit(BReg, value); break;\
	case base+1:	Bit(CReg, value); break;\
	case base+2:	Bit(DReg, value); break;\
	case base+3:	Bit(EReg, value); break;\
	case base+4:	Bit(HReg, value); break;\
	case base+5:	Bit(LReg, value); break;\
	case base+6:	Read8(HL, TempValue); Bit(TempValue, value); break;\
	case base+7:	Bit(AReg, value); break;

#define OpGroupReadNB(base, op)\
	case base:		op(BReg);			break;\
	case base+1:	op(CReg);			break;\
	case base+2:	op(DReg);			break;\
	case base+3:	op(EReg);			break;\
	case base+4:	op(HReg);			break;\
	case base+5:	op(LReg);			break;\
	case base+6:	Read8(HL, TempValue); DoCycles(1); op(TempValue);break;\
	case base+7:	op(AReg);			break;

#define WOpGroupReadNB(base, op)\
	case base:		op(BReg, BBig);			break;\
	case base+1:	op(CReg, CBig);			break;\
	case base+2:	op(DReg, DBig);			break;\
	case base+3:	op(EReg, EBig);			break;\
	case base+4:	op(HReg, HBig);			break;\
	case base+5:	op(LReg, LBig);			break;\
	case base+6:	Read64(HL, TempValue, TempValueBig); DoCycles(1); op(TempValue, TempValueBig);break;\
	case base+7:	op(AReg, ABig);			break;

#define BitwiseGroup(base, op)\
	op(base + 0x00, 0x01, 0x00000000000000ffull);\
	op(base + 0x08, 0x02, 0x000000000000ff00ull);\
	op(base + 0x10, 0x04, 0x0000000000ff0000ull);\
	op(base + 0x18, 0x08, 0x00000000ff000000ull);\
	op(base + 0x20, 0x10, 0x000000ff00000000ull);\
	op(base + 0x28, 0x20, 0x0000ff0000000000ull);\
	op(base + 0x30, 0x40, 0x00ff000000000000ull);\
	op(base + 0x38, 0x80, 0xff00000000000000ull);

#define OpGroupRead(base, op)\
	OpGroupReadNB(base, op)\
	case base+0x46:	Load8n(TempValue); op(TempValue);	break;

#define WOpGroupRead(base, op)\
	WOpGroupReadNB(base, op)\
	case base+0x46:	WLoad8n(TempValue, TempValueBig); op(TempValue, TempValueBig);	break;

#define OpGroupModify(base, op, st)\
	case base:		op(BReg); break;\
	case base+st:	op(CReg); break;\
	case base+2*st:	op(DReg); break;\
	case base+3*st:	op(EReg); break;\
	case base+4*st:	op(HReg); break;\
	case base+5*st:	op(LReg); break;\
	case base+6*st:	Read8(HL, TempValue); DoCycles(1); op(TempValue);	StHL();	break;\
	case base+7*st:	op(AReg); break;

#define WOpGroupModify(base, op, st)\
	case base:		op(BReg, BBig); break;\
	case base+st:	op(CReg, CBig); break;\
	case base+2*st:	op(DReg, DBig); break;\
	case base+3*st:	op(EReg, EBig); break;\
	case base+4*st:	op(HReg, HBig); break;\
	case base+5*st:	op(LReg, LBig); break;\
	case base+6*st:	Read64(HL, TempValue, TempValueBig); DoCycles(1); op(TempValue, TempValueBig);	WStHL();	break;\
	case base+7*st:	op(AReg, ABig); break;


#define CondGroup(base, op)\
	case base+0x00:	op(NZ());	break;\
	case base+0x08:	op(Z());	break;\
	case base+0x10:	op(NC());	break;\
	case base+0x18:	op(C());	break;\
	case base+0x20:	op(PO());	break;\
	case base+0x28:	op(PE());	break;\
	case base+0x30:	op(P());	break;\
	case base+0x38:	op(M());	break;

#define BReg	BC.b.b1
#define CReg	BC.b.b0
#define DReg	DE.b.b1
#define EReg	DE.b.b0
#define HReg	HL.b.b1
#define LReg	HL.b.b0

#define NZ()	Zero
#define	Z()		!Zero
#define	NC()	!Carry
#define C()		Carry
#define PO()	!(Overflow&0x80)
#define PE()	(Overflow&0x80)
#define	P()		!(Sign&0x80)
#define M()		(Sign&0x80)

#define CalculateAddr(ip)\
	Read8(PC, Temp8); PC.a++;\
	DoCycles(2);\
	Addr.a = ip.a + (Sint8)Temp8;

//#define GetT16()		Read8(Addr, TempValue);
#define PutT16()		Write8(Addr, TempValue);
#define WGetT16()		Read64(Addr, TempValue, TempValueBig);
#define WPutT16()		Write64(Addr, TempValue, TempValueBig);
#define IndexedOp(op)	op(TempValue, TempValueBig); WPutT16();
#define ResT16(b, br)	Write64(Addr, TempValue & ~b, TempValueBig & ~br);
#define BitT16(b)		Bit(TempValue, b);
#define SetT16(b, br)	Write64(Addr, TempValue | b, TempValueBig | br);

#define LdIP8(ip, v)\
	CalculateAddr(ip);\
	DoCycles(3);\
	Write8(Addr, v);

#define Ld8IP(ip, v)\
	CalculateAddr(ip);\
	DoCycles(3);\
	Read8(Addr, v);

#define WLdIP8(ip, v, vr)\
	CalculateAddr(ip);\
	DoCycles(3);\
	Write64(Addr, v, vr);

#define WLd8IP(ip, v, vr)\
	CalculateAddr(ip);\
	DoCycles(3);\
	Read64(Addr, v, vr);

#define Loadind(n)	Ld8IP(n, TempValue); DoCycles(1);
#define WLoadind(n)	WLd8IP(n, TempValue, TempValueBig); DoCycles(1);

#define	IndexPage(ip, iph, ipl)\
	Read8(PC, instr);\
	PC.a++;	DoCycles(1);\
	RReg++;\
\
	switch(instr)\
	{\
	default: printf("fd %02x [%04x]\n", instr, PC.a); Halted = true; break;\
\
	case 0xcb :\
		CalculateAddr(ip);\
\
		Read8(PC, instr);\
		PC.a++; DoCycles(1);\
\
		WGetT16();\
		switch(instr)\
		{\
			default: printf("fdcb d %02x\n", instr); Halted = true; break;\
\
			case 0x46: BitT16(0x01); break;\
			case 0x4e: BitT16(0x02); break;\
			case 0x56: BitT16(0x04); break;\
			case 0x5e: BitT16(0x08); break;\
			case 0x66: BitT16(0x10); break;\
			case 0x6e: BitT16(0x20); break;\
			case 0x76: BitT16(0x40); break;\
			case 0x7e: BitT16(0x80); break;\
\
			case 0x86: ResT16(0x01, 0x00000000000000ffull); break;\
			case 0x8e: ResT16(0x02, 0x000000000000ff00ull); break;\
			case 0x96: ResT16(0x04, 0x0000000000ff0000ull); break;\
			case 0x9e: ResT16(0x08, 0x00000000ff000000ull); break;\
			case 0xa6: ResT16(0x10, 0x000000ff00000000ull); break;\
			case 0xae: ResT16(0x20, 0x0000ff0000000000ull); break;\
			case 0xb6: ResT16(0x40, 0x00ff000000000000ull); break;\
			case 0xbe: ResT16(0x80, 0xff00000000000000ull); break;\
\
			case 0xc6: SetT16(0x01, 0x00000000000000ffull); break;\
			case 0xce: SetT16(0x02, 0x000000000000ff00ull); break;\
			case 0xd6: SetT16(0x04, 0x0000000000ff0000ull); break;\
			case 0xde: SetT16(0x08, 0x00000000ff000000ull); break;\
			case 0xe6: SetT16(0x10, 0x000000ff00000000ull); break;\
			case 0xee: SetT16(0x20, 0x0000ff0000000000ull); break;\
			case 0xf6: SetT16(0x40, 0x00ff000000000000ull); break;\
			case 0xfe: SetT16(0x80, 0xff00000000000000ull); break;\
\
			case 0x06: IndexedOp(WRlc);	break;\
			case 0x0e: IndexedOp(WRrc);	break;\
			case 0x16: IndexedOp(WRl);	break;\
			case 0x1e: IndexedOp(WRr);	break;\
			case 0x26: IndexedOp(WSla);	break;\
			case 0x2e: IndexedOp(WSra);	break;\
			case 0x36: IndexedOp(WSll);	break;\
			case 0x3e: IndexedOp(WSrl);	break;\
		}\
		break;\
\
	case 0x24:	Inc8(ip.b.b1);		break;\
	case 0x25:	Dec8(ip.b.b1);		break;\
	case 0x26:	WLoad8n(ip.b.b1, iph);	break;\
	case 0x84:	Add8(ip.b.b1);		break;\
	case 0x8c:	Adc8(ip.b.b1);		break;\
	case 0x94:	Sub8(ip.b.b1);		break;\
	case 0x9c:	Sbc8(ip.b.b1);		break;\
	case 0xa4:	WAnd(ip.b.b1, iph);	break;\
	case 0xac:	WXor(ip.b.b1, iph);	break;\
	case 0xb4:	WOr(ip.b.b1, iph);	break;\
	case 0xbc:	Cp(ip.b.b1);		break;\
\
	case 0x44:	BReg = ip.b.b1;	BBig = iph;	break;\
	case 0x4c:	CReg = ip.b.b1; CBig = iph;	break;\
	case 0x54:	DReg = ip.b.b1; DBig = iph;	break;\
	case 0x5c:	EReg = ip.b.b1; EBig = iph;	break;\
	case 0x7c:	AReg = ip.b.b1; ABig = iph;	break;\
\
	case 0x60:	ip.b.b1 = BReg;	iph = BBig;	break;\
	case 0x61:	ip.b.b1 = CReg; iph = CBig;	break;\
	case 0x62:	ip.b.b1 = DReg; iph = DBig;	break;\
	case 0x63:	ip.b.b1 = EReg; iph = EBig; break;\
	case 0x64:						break;\
	case 0x65:	ip.b.b1 = ip.b.b0; iph = ipl; break;\
	case 0x67:	ip.b.b1 = AReg; iph = ABig;	break;\
\
	case 0x2c:	Inc8(ip.b.b0);		break;	/* inc ipl */\
	case 0x2d:	Dec8(ip.b.b0);		break;	/* dec ipl */\
	case 0x2e:	WLoad8n(ip.b.b0, ipl);	break;	/* ld ipl, n*/\
	case 0x85:	Add8(ip.b.b0);		break;	/* add ipl */\
	case 0x8d:	Adc8(ip.b.b0);		break;	/* adc ipl */\
	case 0x95:	Sub8(ip.b.b0);		break;	/* sub ipl */\
	case 0x9d:	Sbc8(ip.b.b0);		break;	/* sbc ipl */\
	case 0xa5:	WAnd(ip.b.b0, ipl);	break;	/* and ipl */\
	case 0xad:	WXor(ip.b.b0, ipl);	break;	/* xor ipl */\
	case 0xb5:	WOr(ip.b.b0, ipl);	break;	/* or ipl */\
	case 0xbd:	Cp(ip.b.b0);		break;	/* cp ipl */\
\
	case 0x45:	BReg = ip.b.b0; BBig = ipl;	break;\
	case 0x4d:	CReg = ip.b.b0;	CBig = ipl;	break;\
	case 0x55:	DReg = ip.b.b0;	DBig = ipl;	break;\
	case 0x5d:	EReg = ip.b.b0;	EBig = ipl;	break;\
	case 0x7d:	AReg = ip.b.b0;	ABig = ipl;	break;\
\
	case 0x68:	ip.b.b0 = BReg;	ipl = BBig;	break;\
	case 0x69:	ip.b.b0 = CReg;	ipl = CBig;	break;\
	case 0x6a:	ip.b.b0 = DReg;	ipl = DBig;	break;\
	case 0x6b:	ip.b.b0 = EReg;	ipl = EBig;	break;\
	case 0x6c:	ip.b.b0 = ip.b.b1; ipl = iph;	break;\
	case 0x6d:						break;\
	case 0x6f:	ip.b.b0 = AReg;	ipl = ABig;	break;\
\
	case 0x23:	Inc16(ip);			break;\
	case 0x2b:	Dec16(ip);			break;\
\
	case 0x34:	Loadind(ip); Inc8(TempValue); PutT16();	break;\
	case 0x35:	Loadind(ip); Dec8(TempValue); PutT16();	break;\
\
	case 0x09:	Add16(ip, BC);	break;\
	case 0x19:	Add16(ip, DE);	break;\
	case 0x29:	Add16(ip, ip);	break;\
	case 0x39:	Add16(ip, SP);	break;\
\
	case 0x21:	WLoad16nn(ip, iph, ipl);	break;\
	case 0x22:	WLoadnn16(ip, iph, ipl);	break;\
	case 0x2a:	WLoad16nni(ip, iph, ipl);	break;\
\
	case 0x86:	Loadind(ip); Add8(TempValue);	break;\
	case 0x8e:	Loadind(ip); Adc8(TempValue);	break;\
	case 0x96:	Loadind(ip); Sub8(TempValue);	break;\
	case 0x9e:	Loadind(ip); Sbc8(TempValue);	break;\
	case 0xa6:	WLoadind(ip); WAnd(TempValue, TempValueBig);	break;\
	case 0xae:	WLoadind(ip); WXor(TempValue, TempValueBig);	break;\
	case 0xb6:	WLoadind(ip); WOr(TempValue, TempValueBig);		break;\
	case 0xbe:	Loadind(ip); Cp(TempValue);		break;\
\
	case 0x70:	WLdIP8(ip, BReg, BBig);	break;\
	case 0x71:	WLdIP8(ip, CReg, CBig);	break;\
	case 0x72:	WLdIP8(ip, DReg, DBig);	break;\
	case 0x73:	WLdIP8(ip, EReg, EBig);	break;\
	case 0x74:	WLdIP8(ip, HReg, HBig);	break;\
	case 0x75:	WLdIP8(ip, LReg, LBig);	break;\
	case 0x77:	WLdIP8(ip, AReg, ABig);	break;\
\
	case 0x46:	WLd8IP(ip, BReg, BBig);	break;\
	case 0x4e:	WLd8IP(ip, CReg, CBig);	break;\
	case 0x56:	WLd8IP(ip, DReg, DBig);	break;\
	case 0x5e:	WLd8IP(ip, EReg, EBig);	break;\
	case 0x66:	WLd8IP(ip, HReg, HBig);	break;\
	case 0x6e:	WLd8IP(ip, LReg, LBig);	break;\
	case 0x7e:	WLd8IP(ip, AReg, ABig);	break;\
\
	case 0x36:	\
		CalculateAddr(ip);\
		Read64(PC, TempValue, TempValueBig); PC.a++;\
		WPutT16();\
	break;	/* ld (ip+d), n */\
\
	case 0xe5:	WPush16Instr(ip, iph, ipl);	break;	/* push ip */\
	case 0xe1:	WPop16(ip, iph, ipl);		break;	/* pop ip */\
	case 0xe3: /* ex (sp), ip */\
		WPop16(Temp16, Temp64, TempValueBig); WPush16(ip, iph, ipl);\
		ip.a = Temp16.a;\
		iph = Temp64;\
		ipl = TempValueBig;\
	break;\
	case 0xe9:	PC.a = ip.a;	break;	/* jp (ip) */\
	case 0xf9:	Load1616(SP, ip);	break;	/* ld sp, ip */\
	}

#define LdGroup(base, r, rw)\
	case base:		r = BReg;	rw = BBig;	break;\
	case base+1:	r = CReg;	rw = CBig;	break;\
	case base+2:	r = DReg;	rw = DBig;	break;\
	case base+3:	r = EReg;	rw = EBig;	break;\
	case base+4:	r = HReg;	rw = HBig;	break;\
	case base+5:	r = LReg;	rw = LBig;	break;\
	case base+6:	Read64(HL, r, rw);		break;\
	case base+7:	r = AReg;	rw = ABig;	break;

#define LargeRegisterGroup(base, op)\
	case base:		op(BC);	break;\
	case base+0x10:	op(DE);	break;\
	case base+0x20:	op(HL);	break;\
	case base+0x30:	op(SP);	break;

#define WLargeRegisterGroup(base, op, ops)\
	case base:		op(BC, BBig, CBig);	break;\
	case base+0x10:	op(DE, DBig, EBig);	break;\
	case base+0x20:	op(HL, HBig, LBig);	break;\
	case base+0x30:	ops(SP);			break;

#define Add16HL(v) Add16(HL, v); DoCycles(4);

int CZ80::CPUThreadHelper(void *tptr)
{
	return ((CZ80 *)tptr)->CPUThread();
}

int CZ80::CPUThread()
{
	bool Halted = false;
	Uint8 LastInstr = 0;

	/* wait to be told to go for the first time */
	SDL_SemWait(GoSemaphore);

	while(!Quit)
	{
		RegPair Addr;
		Uint8 Temp8, TempValue;
		Uint64 Temp64, TempValueBig;
		RegPair Temp16;
		Uint32 Temp32;
		Uint8 instr;

		/* check for interrupt */
//		static bool Dump = false;
		if(ILine && IFF2 && LastInstr != 0xfb)
		{
//			Dump = true;
			if(Halted)
			{
				Halted = false;
				PC.a++;
			}
			IFF2 = IFF1 = false;
			SubCycleCount += 6 + (IMode ? 1 : 0);
			Push16(PC);
			switch(IMode)
			{
				case 0:
				case 1:	PC.a = 0x0038;	break;
				case 2:
					RegPair Addr;
					Addr.b.b1 = IReg;
					Addr.b.b0 = 0xff;

					Read8(Addr, PC.b.b0); Addr.a++;
					Read8(Addr, PC.b.b1);
				break;
			}
		}

		/* check for trap */
		if(TrapFunc && CatchTab[PC.a >> 5] & (1 << (PC.a&31)))
			TrapFunc(TrapParam, *this, PC.a);

		/* fetch instruction */
/*		if(CarryBig&~0xff) printf("Carrybig error at %04x - %02x\n", PC.a, LastInstr);
		if(Dump)
		{
			static int LastTotal = 0;
#define Read8Free(n)		Read8Ptrs[((n)&0xffff) >> 14][((n)&0xffff)&16383]
			printf("%04x %02x %02x %02x %02x: %d\n", PC.a, Read8Free(PC.a),  Read8Free(PC.a+1),  Read8Free(PC.a+2),  Read8Free(PC.a+3), (SubCycleCount + TotalCycleCount) - LastTotal);
			LastTotal = (SubCycleCount + TotalCycleCount);
		}*/
		Read8(PC, instr); LastInstr = instr;
		PC.a++; DoCycles(1);
		RReg++;

		switch(instr)
		{
			default: printf("%02x [%04x]\n", instr, PC.a); Halted = true; break;

			case 0x00:					break; //NOP

			case 0x36:	
				WLoad8n(TempValue, TempValueBig);
				WStHL();
			break; //LD (HL), n

			WLargeRegisterGroup(0x01, WLoad16nn, Load16nn);
			LargeRegisterGroup(0x03, Inc16);
			LargeRegisterGroup(0x09, Add16HL);
			LargeRegisterGroup(0x0b, Dec16);

			LdGroup(0x40, BReg, BBig);
			LdGroup(0x48, CReg, CBig);
			LdGroup(0x50, DReg, DBig);
			LdGroup(0x58, EReg, EBig);
			LdGroup(0x60, HReg, HBig);
			LdGroup(0x68, LReg, LBig);
			LdGroup(0x78, AReg, ABig);

			case 0x70:	Write64(HL, BReg, BBig);	break; //LD (HL), B
			case 0x71:	Write64(HL, CReg, CBig);	break; //LD (HL), C
			case 0x72:	Write64(HL, DReg, DBig);	break; //LD (HL), D
			case 0x73:	Write64(HL, EReg, EBig);	break; //LD (HL), E
			case 0x74:	Write64(HL, HReg, HBig);	break; //LD (HL), H
			case 0x75:	Write64(HL, LReg, LBig);	break; //LD (HL), L
			case 0x77:	Write64(HL, AReg, ABig);	break; //LD (HL), A

			case 0x06:	WLoad8n(BReg, BBig);		break; //LD B, n
			case 0x0e:	WLoad8n(CReg, CBig);		break; //LD C, n
			case 0x16:	WLoad8n(DReg, DBig);		break; //LD D, n
			case 0x1e:	WLoad8n(EReg, EBig);		break; //LD E, n
			case 0x26:	WLoad8n(HReg, HBig);		break; //LD H, n
			case 0x2e:	WLoad8n(LReg, LBig);		break; //LD L, n
			case 0x3e:	WLoad8n(AReg, ABig);		break; //LD A, n

			case 0x0a:	Read64(BC, AReg, ABig);		break; //LD A, (BC)
			case 0x02:	Write64(BC, AReg, ABig);	break; //LD (BC), A
			case 0x1a:	Read64(DE, AReg, ABig);		break; //LD A, (DE)
			case 0x12:	Write64(DE, AReg, ABig);	break; //LD (DE), A

			OpGroupRead(0x80, Add8);
			OpGroupRead(0x88, Adc8);
			OpGroupRead(0x90, Sub8);
			OpGroupRead(0x98, Sbc8);
			WOpGroupRead(0xa0, WAnd);
			WOpGroupRead(0xa8, WXor);
			WOpGroupRead(0xb0, WOr);
			OpGroupRead(0xb8, Cp);

			OpGroupModify(0x05, Dec8, 8);
			OpGroupModify(0x04, Inc8, 8);

			case 0x10:	BReg--; DoCycles(1); JR(BReg); break; //DJNZ

			case 0x18:	JR(1);		break; //JR
			case 0x20:	JR(NZ());	break; //JR NZ
			case 0x28:	JR(Z());	break; //JR Z
			case 0x30:	JR(NC());	break; //JR NC
			case 0x38:	JR(C());	break; //JR C

			CondGroup(0xc0, Ret);
			CondGroup(0xc2, JP);
			CondGroup(0xc4, Call);

			case 0xc3:	JP(1);		break; //JP
			case 0xc9:	Pop16(PC);	break; //RET
			case 0xcd:	Call(1);	break; //CALL

			case 0xf3:	IFF1 = IFF2 = false;	break; //DI
			case 0xfb:	IFF1 = IFF2 = true;		break; //EI

			case 0x22:	WLoadnn16(HL, HBig, LBig);		break; //LD (nn), HL
			case 0x32:
				Load16nn(Addr);
				Write64(Addr, AReg, ABig);
			break; //LD (nn), A
			case 0x2a:	WLoad16nni(HL, HBig, LBig);		break; //LD HL, (nn)
			case 0x3a:	WLoad8nn(AReg, ABig);	break; //LD A, (nn)

			case 0xf9:	Load1616(SP, HL);	break; //LD SP, HL

			case 0xc5:	WPush16Instr(BC, BBig, CBig);			break; //PUSH BC
			case 0xd5:	WPush16Instr(DE, DBig, EBig);			break; //PUSH DE
			case 0xe5:	WPush16Instr(HL, HBig, LBig);			break; //PUSH HL
			case 0xf5:	ReadFromAF(Temp16); WPush16Instr(Temp16, ABig, CarryBig);	break; //PUSH AF

			case 0xc1:	WPop16(BC, BBig, CBig);				break; //POP BC
			case 0xd1:	WPop16(DE, DBig, EBig);				break; //POP DE
			case 0xe1:	WPop16(HL, HBig, LBig);				break; //POP HL
			case 0xf1:	WPop16(Temp16, ABig, CarryBig); CarryBig &= 0xff; WriteToAF(Temp16); 	break; //POP AF

			case 0xd3:
				Load8n(Addr.b.b0);
				Addr.b.b1 = AReg;
				IOAccess(Addr.a);
				WritePort(Addr.a, AReg, TotalCycleCount+SubCycleCount);
				DoCycles(4);
			break; //OUT (n), A

			case 0xdb:
				Load8n(Addr.b.b0);
				Addr.b.b1 = AReg;
				IOAccess(Addr.a);
				ReadPort(Addr.a, AReg, TotalCycleCount+SubCycleCount);
				DoCycles(4);
			break; //IN A, (n)

			case 0xd9:
				BBig2 ^= BBig; BBig ^= BBig2; BBig2 ^= BBig;
				CBig2 ^= CBig; CBig ^= CBig2; CBig2 ^= CBig;
				DBig2 ^= DBig; DBig ^= DBig2; DBig2 ^= DBig;
				EBig2 ^= EBig; EBig ^= EBig2; EBig2 ^= EBig;
				HBig2 ^= HBig; HBig ^= HBig2; HBig2 ^= HBig;
				LBig2 ^= LBig; LBig ^= LBig2; LBig2 ^= LBig;

				BC.a ^= BC2.a; BC2.a ^= BC.a; BC.a ^= BC2.a;
				HL.a ^= HL2.a; HL2.a ^= HL.a; HL.a ^= HL2.a;
				DE.a ^= DE2.a; DE2.a ^= DE.a; DE.a ^= DE2.a;
			break; //EXX

			case 0xeb:
				DBig ^= HBig; HBig ^= DBig; DBig ^= HBig;
				EBig ^= LBig; LBig ^= EBig; EBig ^= LBig;

				DE.a ^= HL.a; HL.a ^= DE.a; DE.a ^= HL.a;
			break; //EX DE, HL

			case 0xe3:
				WPop16(Temp16, Temp64, TempValueBig); WPush16(HL, HBig, LBig);
				
				HL.a = Temp16.a;
				HBig = Temp64;
				LBig = TempValueBig;
			break; //EX (SP), HL

			case 0x08:
				ReadFromAF(Temp16);
				WriteToAF(AF2);
				AF2.a = Temp16.a;

				ABig ^= ABig2; ABig2 ^= ABig; ABig ^= ABig2;
				CarryBig ^= CarryBig2; CarryBig2 ^= CarryBig; CarryBig ^= CarryBig2;
			break; //EX AF, AF'

			case 0xe9:	PC.a = HL.a;	break; //JP (HL)

			case 0x0f:
				Carry = AReg&1;
				Misc = AReg = (AReg >> 1) | (Carry << 7);
				LastOp = false;
				HalfCarry = 0;

				CarryBig = ABig&0xff;
				ABig = (ABig >> 8) | (CarryBig << 56);
			break; //RRCA

			case 0x1f:
				Temp8 = Carry;
				Carry = AReg&1;
				Misc = AReg = (AReg >> 1) | (Temp8 << 7);
				LastOp = false;
				HalfCarry = 0;

				Temp64 = CarryBig;
				CarryBig = ABig&0xff;
				ABig = (ABig >> 8) | (Temp64 << 56);
			break; //RRA

			case 0x07:
				Carry = AReg >> 7;
				Misc = AReg = (AReg << 1) | Carry;
				LastOp = false;
				HalfCarry = 0;

				CarryBig = ABig >> 56;
				ABig = (ABig << 8) | CarryBig;
			break; //RLCA

			case 0x17:
				Temp8 = Carry;
				Carry = AReg >> 7;
				Misc = AReg = (AReg << 1) | Temp8;
				LastOp = false;
				HalfCarry = 0;

				Temp64 = CarryBig;
				CarryBig = ABig >> 56;
				ABig = (ABig << 8) | Temp64;
			break; //RLA

			case 0xc7:	Push16(PC);	PC.a = 0x00;	break; //RST 00h
			case 0xcf:	Push16(PC);	PC.a = 0x08;	break; //RST 08h
			case 0xd7:	Push16(PC);	PC.a = 0x10;	break; //RST 10h
			case 0xdf:	Push16(PC);	PC.a = 0x18;	break; //RST 18h
			case 0xe7:	Push16(PC);	PC.a = 0x20;	break; //RST 20h
			case 0xef:	Push16(PC);	PC.a = 0x28;	break; //RST 28h
			case 0xf7:	Push16(PC);	PC.a = 0x30;	break; //RST 30h
			case 0xff:	Push16(PC);	PC.a = 0x38;	break; //RST 38h

			case 0x2f:
				AReg ^= 0xff;
				ABig ^= 0xffffffffffffffffull;
				Misc = AReg;
				HalfCarry = 0x10;
				LastOp = true;
			break; //CPL

			case 0x27:
				TempValue =
					((HalfCarry || ((AReg&0xf) > 9)) ? 0x06 : 0x00) |
					((Carry || (AReg > 0x9f)) ? 0x60 : 0x00);

				if(LastOp)
				{
					/* sub8 without overflow calculation, and with carry adjustment */
					Temp16.a = AReg - TempValue;
					HalfCarry = ((AReg&0xf) - (TempValue&0xf))&0x10;
					Carry |= Temp16.b.b1&1;
					Misc = Sign = Zero = AReg = Temp16.b.b0;
				}
				else
				{
					if((AReg > 0x90) && ((AReg&0xf) > 9)) TempValue |= 0x60;

					/* add8 without overflow calculation, and with carry adjustment */
					Temp16.a = AReg + TempValue;
					HalfCarry = ((AReg&0xf) + (TempValue&0xf))&0x10;
					Carry |= Temp16.b.b1;
					Misc = Sign = Zero = AReg = Temp16.b.b0;
				}

				Overflow = ParityTab[AReg];
			break; //DAA

			case 0x76:
				PC.a--; Halted = true;
				DoCycles(CyclesToRun - SubCycleCount);
			break; //HALT

			case 0x37: Misc = AReg; Carry = 1; LastOp = false; HalfCarry = 0; CarryBig = 0xff; break; //SCF
			case 0x3f:
				Misc = AReg; LastOp = false;
				HalfCarry = Carry << 4; Carry ^= 1;
				CarryBig ^= 0xff;
			break; //CCF

			case 0xdd:	IndexPage(IX, IXhBig, IXlBig);	break; /* DD */
			case 0xfd:	IndexPage(IY, IYhBig, IYlBig);	break; /* FD */

			/* CB */
			case 0xcb:
				Read8(PC, instr);
				PC.a++;
				DoCycles(1);
				RReg++;

				switch(instr)
				{
					WOpGroupModify(0x00, WRlc, 1);
					WOpGroupModify(0x08, WRrc, 1);

					WOpGroupModify(0x10, WRl, 1);
					WOpGroupModify(0x18, WRr, 1);

					WOpGroupModify(0x20, WSla, 1);
					WOpGroupModify(0x28, WSra, 1);

					WOpGroupModify(0x30, WSll, 1);
					WOpGroupModify(0x38, WSrl, 1);

					BitwiseGroup(0x40, BitGroup);
					BitwiseGroup(0x80, ResGroup);
					BitwiseGroup(0xc0, SetGroup);
				}
			break;

			/* ED */
			case 0xed:
				Read8(PC, instr);
				PC.a++;
				DoCycles(1);
				RReg++;

				switch(instr)
				{
					default: printf("ed %02x [%04x]\n", instr, PC.a); Halted = true; break;

					case 0x46: case 0x4e: case 0x66: case 0x6e: IMode = 0; break; //IM 0
					case 0x56: case 0x76: IMode = 1; break; //IM 1
					case 0x5e: case 0x7e: IMode = 2; break; //IM 2

					case 0x41: Out(BC.a, BReg);	break; //OUT (C), B
					case 0x49: Out(BC.a, CReg);	break; //OUT (C), C
					case 0x51: Out(BC.a, DReg);	break; //OUT (C), D
					case 0x59: Out(BC.a, EReg);	break; //OUT (C), E
					case 0x61: Out(BC.a, HReg);	break; //OUT (C), H
					case 0x69: Out(BC.a, LReg);	break; //OUT (C), L
					case 0x79: Out(BC.a, AReg);	break; //OUT (C), A

					case 0x71: Out(BC.a, 0);	break; //OUT (C), 0

					case 0x40: In(BC.a, BReg);	break; //IN B, (C)
					case 0x48: In(BC.a, CReg);	break; //IN C, (C)
					case 0x50: In(BC.a, DReg);	break; //IN D, (C)
					case 0x58: In(BC.a, EReg);	break; //IN E, (C)
					case 0x60: In(BC.a, HReg);	break; //IN H, (C)
					case 0x68: In(BC.a, LReg);	break; //IN L, (C)
					case 0x78: In(BC.a, AReg);	break; //IN A, (C)

					case 0x70: In(BC.a, Temp8);	break; //IN F, (C)

					case 0x44: case 0x4c: case 0x54: case 0x5c:
					case 0x64: case 0x6c: case 0x74: case 0x7c:
						TempValue = AReg;
						AReg = 0;
						Sub8(TempValue);
					break; //NEG

					case 0x47:	IReg = AReg; DoCycles(1); break; //LD I, A
					case 0x57:
						Misc = Sign = Zero = AReg = IReg;
						LastOp = false;
						Overflow = IFF2 ? 0x80 : 0;
						HalfCarry = 0;
						DoCycles(1);
					break; //LD A, I

					case 0x4f:
						RReg = AReg;
						RHigh = AReg&0x80;
						DoCycles(1);
					break; //LD R, A
					case 0x5f:
						Misc = Sign = Zero = AReg = (RReg&0x7f) | RHigh;
						LastOp = false;
						Overflow = IFF2 ? 0x80 : 0;
						HalfCarry = 0;
						DoCycles(1);
					break; //LD A, R

					LargeRegisterGroup(0x42, Sbc16);
					WLargeRegisterGroup(0x43, WLoadnn16, Loadnn16);
					LargeRegisterGroup(0x4a, Adc16);
					WLargeRegisterGroup(0x4b, WLoad16nni, Load16nni);

					case 0xb8:
						CopyHLToDE();
						DE.a--;	HL.a--;
						DoCycles(2);
						if(BC.a) {PC.a -= 2; DoCycles(5);}
					break; //LDDR

					case 0xb0:
						CopyHLToDE();
						DE.a++;	HL.a++;
						DoCycles(2);
						if(BC.a) {PC.a -= 2; DoCycles(5);}
					break; //LDIR

					case 0xa0:
						CopyHLToDE();
						DE.a++;	HL.a++;
						DoCycles(2);
					break; //LDI

					case 0xa8:
						CopyHLToDE();
						DE.a--;	HL.a--;
						DoCycles(2);
					break; //LDD


					case 0xa1: CPICompare(); HL.a++; break; //CPI
					case 0xa9: CPICompare(); HL.a--; break; //CPD

					case 0xb1:
						CPICompare();
						HL.a++;
						if(BC.a && Zero) {PC.a -= 2; DoCycles(5);}
					break; //CPIR

					case 0xb9:
						CPICompare();
						HL.a--;
						if(BC.a && Zero) {PC.a -= 2; DoCycles(5);}
					break; //CPDR


					case 0xab:
						Read8(HL, TempValue); HL.a--;
						Out(BC.a, TempValue);
						Dec8(BReg);
					break; //OUTD

					case 0xa3:
						Read8(HL, TempValue); HL.a++;
						Out(BC.a, TempValue);
						Inc8(BReg);
					break; //OUTI


					case 0x6f:
						Read64(HL, Temp16.b.b0, Temp64);

						TempValue = (Temp16.b.b0 << 4) | (AReg&0x0f);
						AReg = (AReg&0xf0) | (Temp16.b.b0 >> 4);

						TempValueBig = (Temp64 << 32) | (ABig&0xffffffff);
						ABig = (ABig&0xffffffff00000000ull) | (Temp64 >> 32);

						WStHL();

						Overflow = ParityTab[Sign = Zero = Misc = AReg];
						LastOp = false; HalfCarry = 0;
					break; //RLD

					case 0x67:
						Read64(HL, Temp16.b.b0, Temp64);

						TempValue = (Temp16.b.b0 >> 4) | (AReg << 4);
						AReg = (AReg&0xf0) | (Temp16.b.b0&0x0f);

						TempValueBig = (Temp64 >> 32) | (ABig << 32);
						ABig = (ABig&0xffffffff00000000ull) | (Temp64&0xffffffffull);

						WStHL();

						Overflow = ParityTab[Sign = Zero = Misc = AReg];
						LastOp = false; HalfCarry = 0;
					break; //RRD

					case 0x45: case 0x4d: case 0x55: case 0x5d:
					case 0x65: case 0x6d: case 0x75: case 0x7d:
						IFF1 = IFF2;
						Ret(1);
					break; //RETN
				}
			break;
		}
	}

	SDL_SemPost(StopSemaphore);
	return 0;
}

RegisterState CZ80::GetState()
{
	RegisterState rs;

	rs.IFF1 = IFF1; rs.IFF2 = IFF2;
	rs.IMode = IMode; rs.RReg = RReg | RHigh;
	rs.IReg = IReg;
	ReadFromAF(rs.AF);
	rs.BC = BC; rs.DE = DE; rs.HL = HL; rs.SP = SP;
	rs.BC2 = BC2; rs.DE2 = DE2; rs.HL2 = HL2; rs.AF2 = AF2;
	rs.IY = IY; rs.IX = IX; rs.PC = PC;

	return rs;
}

void CZ80::SetState(RegisterState rs)
{
	IFF1 = rs.IFF1; IFF2 = rs.IFF2;
	IMode = rs.IMode; RReg = rs.RReg&127; RHigh = rs.RReg&128;
	IReg = rs.IReg;
	WriteToAF(rs.AF);
	BC = rs.BC; DE = rs.DE; HL = rs.HL; SP = rs.SP;
	BC2 = rs.BC2; DE2 = rs.DE2; HL2 = rs.HL2; AF2 = rs.AF2;
	IY = rs.IY; IX = rs.IX; PC = rs.PC;
}
